# pylint: disable=too-many-lines
import itertools
from abc import abstractmethod
from collections import OrderedDict
from dataclasses import dataclass, field as dataclass_field
from enum import Enum
from sys import intern
from typing import List, Mapping, Optional, Sequence, Tuple, Union

import rflx.identifier
from rflx.common import Base, file_name, indent, indent_next, unique
from rflx.contract import invariant


class ID(rflx.identifier.ID):
    @property
    def _separator(self) -> str:
        return "."

    def __add__(self, other: object) -> "ID":
        result = super().__add__(other)
        assert isinstance(result, ID)
        return result

    def __radd__(self, other: object) -> "ID":
        result = super().__radd__(other)
        assert isinstance(result, ID)
        return result

    def __mul__(self, other: object) -> "ID":
        result = super().__mul__(other)
        assert isinstance(result, ID)
        return result

    def __rmul__(self, other: object) -> "ID":
        result = super().__rmul__(other)
        assert isinstance(result, ID)
        return result


StrID = Union[str, ID]


class Precedence(Enum):
    UNDEFINED = 0
    BOOLEAN_OPERATOR = 1
    RELATIONAL_OPERATOR = 2
    BINARY_ADDING_OPERATOR = 3
    UNARY_ADDING_OPERATOR = 4
    MULTIPLYING_OPERATOR = 5
    HIGHEST_PRECEDENCE_OPERATOR = 6
    LITERAL = 7


class Expr(Base):
    _str: str

    def __str__(self) -> str:
        try:
            return self._str
        except AttributeError:
            self._update_str()
            return self._str

    def __neg__(self) -> "Expr":
        raise NotImplementedError

    @abstractmethod
    def _update_str(self) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def precedence(self) -> Precedence:
        raise NotImplementedError

    def parenthesized(self, expr: "Expr") -> str:
        if expr.precedence.value <= self.precedence.value:
            return "(" + indent_next(str(expr), 1) + ")"
        return str(expr)


class Not(Expr):
    def __init__(self, expr: Expr) -> None:
        super().__init__()
        self.expr = expr

    def _update_str(self) -> None:
        self._str = intern(f"not {self.parenthesized(self.expr)}")

    @property
    def precedence(self) -> Precedence:
        return Precedence.HIGHEST_PRECEDENCE_OPERATOR


class BinExpr(Expr):
    def __init__(self, left: Expr, right: Expr) -> None:
        super().__init__()
        self.left = left
        self.right = right

    def _update_str(self) -> None:
        self._str = intern(
            f"{self.parenthesized(self.left)}{self.symbol}{self.parenthesized(self.right)}"
        )

    @property
    @abstractmethod
    def precedence(self) -> Precedence:
        raise NotImplementedError

    @property
    @abstractmethod
    def symbol(self) -> str:
        raise NotImplementedError


class AssExpr(Expr):
    def __init__(self, *terms: Expr) -> None:
        super().__init__()
        self.terms = list(terms)

    def _update_str(self) -> None:
        self._str = intern(self.symbol.join(map(self.parenthesized, self.terms)))

    @property
    @abstractmethod
    def precedence(self) -> Precedence:
        raise NotImplementedError

    @property
    @abstractmethod
    def symbol(self) -> str:
        raise NotImplementedError


class BoolAssExpr(AssExpr):
    def _update_str(self) -> None:
        if not self.terms:
            self._str = str(TRUE)
            return
        self._str = ""
        for i, t in reversed(list(enumerate(self.terms))):
            if i == 0:
                self._str = self.parenthesized(t) + self._str
            else:
                self._str = (
                    "\n"
                    + str(self.symbol)[1:]
                    + indent_next(self.parenthesized(t), len(self.symbol) - 1)
                    + self._str
                )
        self._str = intern(self._str)

    @property
    @abstractmethod
    def symbol(self) -> str:
        raise NotImplementedError


class And(BoolAssExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.BOOLEAN_OPERATOR

    @property
    def symbol(self) -> str:
        return " and "


class AndThen(And):
    @property
    def symbol(self) -> str:
        return " and then "


class Or(BoolAssExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.BOOLEAN_OPERATOR

    @property
    def symbol(self) -> str:
        return " or "


class OrElse(Or):
    @property
    def symbol(self) -> str:
        return " or else "


class Number(Expr):
    def __init__(self, value: int, base: int = 0) -> None:
        super().__init__()
        self.value = value
        self.base = base

    def __neg__(self) -> "Number":
        return Number(-self.value)

    def _update_str(self) -> None:
        value = self.value if self.value >= 0 else -self.value
        if self.base == 0:
            self._str = "{}".format(value)
        elif self.base == 2:
            self._str = "2#{:b}#".format(value)
        elif self.base == 8:
            self._str = "8#{:o}#".format(value)
        elif self.base == 10:
            self._str = "10#{}#".format(value)
        elif self.base == 16:
            self._str = "16#{:X}#".format(value)
        else:
            raise NotImplementedError(f"unsupported base {self.base}")
        self._str = intern(f"(-{self._str})" if self.value < 0 else self._str)

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class Add(AssExpr):
    def _update_str(self) -> None:
        self._str = str(self.terms[0])
        for t in self.terms[1:]:
            if (isinstance(t, Number) and t.value < 0) or (isinstance(t, Name) and t.negative):
                self._str += f" - {self.parenthesized(-t)}"
            else:
                self._str += f"{self.symbol}{self.parenthesized(t)}"
        self._str = intern(self._str)

    @property
    def precedence(self) -> Precedence:
        return Precedence.BINARY_ADDING_OPERATOR

    @property
    def symbol(self) -> str:
        return " + "


class Concatenation(AssExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.BINARY_ADDING_OPERATOR

    @property
    def symbol(self) -> str:
        return " & "


class Mul(AssExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.MULTIPLYING_OPERATOR

    @property
    def symbol(self) -> str:
        return " * "


class Sub(BinExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.BINARY_ADDING_OPERATOR

    @property
    def symbol(self) -> str:
        return " - "


class Div(BinExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.MULTIPLYING_OPERATOR

    @property
    def symbol(self) -> str:
        return " / "


class Pow(BinExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.HIGHEST_PRECEDENCE_OPERATOR

    @property
    def symbol(self) -> str:
        return "**"


class Mod(BinExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.MULTIPLYING_OPERATOR

    @property
    def symbol(self) -> str:
        return " mod "


class Name(Expr):
    def __init__(self, negative: bool = False) -> None:
        super().__init__()
        self.negative = negative
        self._update_str()

    def _update_str(self) -> None:
        self._str = intern(f"(-{self._representation})" if self.negative else self._representation)

    @property
    @abstractmethod
    def _representation(self) -> str:
        raise NotImplementedError

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class Variable(Name):
    def __init__(self, identifier: StrID, negative: bool = False) -> None:
        self.identifier = ID(identifier)
        super().__init__(negative)

    def __neg__(self) -> "Variable":
        return self.__class__(self.identifier, not self.negative)

    @property
    def _representation(self) -> str:
        return str(self.name)

    @property
    def name(self) -> str:
        return str(self.identifier)


TRUE = Variable("True")
FALSE = Variable("False")


class Attribute(Name):
    def __init__(self, prefix: Union[StrID, Expr], negative: bool = False) -> None:
        if isinstance(prefix, ID):
            prefix = Variable(prefix)
        if isinstance(prefix, str):
            prefix = Variable(prefix)

        self.prefix: Expr = prefix
        super().__init__(negative)

    def __neg__(self) -> "Attribute":
        return self.__class__(self.prefix, not self.negative)

    @property
    def _representation(self) -> str:
        return f"{self.prefix}'{self.__class__.__name__}"


class Size(Attribute):
    pass


class Length(Attribute):
    pass


class First(Attribute):
    pass


class Last(Attribute):
    pass


class Range(Attribute):
    pass


class Old(Attribute):
    pass


class Result(Attribute):
    pass


class Constrained(Attribute):
    pass


class Valid(Attribute):
    pass


class AttributeExpr(Attribute):
    def __init__(
        self, prefix: Union[StrID, Expr], expression: Expr, negative: bool = False
    ) -> None:
        self.expression = expression
        super().__init__(prefix)

    @property
    def _representation(self) -> str:
        return f"{self.prefix}'{self.__class__.__name__} ({self.expression})"


class Val(AttributeExpr):
    pass


class Pos(AttributeExpr):
    pass


@invariant(lambda self: len(self.elements) > 0)
class Indexed(Name):
    def __init__(self, prefix: Expr, *elements: Expr, negative: bool = False) -> None:
        self.prefix = prefix
        self.elements = list(elements)
        super().__init__(negative)

    def __neg__(self) -> "Indexed":
        return self.__class__(self.prefix, *self.elements, negative=not self.negative)

    @property
    def _representation(self) -> str:
        return f"{self.prefix} (" + ", ".join(map(str, self.elements)) + ")"


class Selected(Name):
    def __init__(
        self,
        prefix: Expr,
        selector_identifier: StrID,
        negative: bool = False,
    ) -> None:
        self.prefix = prefix
        self.selector_identifier = ID(selector_identifier)
        super().__init__(negative)

    def __neg__(self) -> "Selected":
        return self.__class__(self.prefix, self.selector_identifier, not self.negative)

    @property
    def _representation(self) -> str:
        return f"{self.prefix}.{self.selector_identifier}"


class Call(Name):
    def __init__(
        self,
        identifier: StrID,
        args: Sequence[Expr] = None,
        negative: bool = False,
    ) -> None:
        self.identifier = ID(identifier)
        self.args = args or []
        super().__init__(negative)

    def __neg__(self) -> "Call":
        return self.__class__(self.identifier, self.args, not self.negative)

    @property
    def _representation(self) -> str:
        args = ", ".join(map(str, self.args))
        if args:
            args = f" ({args})"
        call = f"{self.identifier}{args}"
        return call


class Slice(Name):
    def __init__(self, prefix: Expr, first: Expr, last: Expr) -> None:
        self.prefix = prefix
        self.first = first
        self.last = last
        super().__init__()

    @property
    def _representation(self) -> str:
        return f"{self.prefix} ({self.first} .. {self.last})"


class Aggregate(Expr):
    def __init__(self, *elements: Expr) -> None:
        super().__init__()
        self.elements = list(elements)

    def _update_str(self) -> None:
        assert len(self.elements) > 1
        self._str = intern("(" + ", ".join(map(str, self.elements)) + ")")

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class String(Aggregate):
    def __init__(self, data: str) -> None:
        data = data.replace('"', '""')
        super().__init__(*[Number(ord(d)) for d in data])
        self.data = data

    def _update_str(self) -> None:
        self._str = intern(f'"{self.data}"')

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class NamedAggregate(Expr):
    def __init__(self, *elements: Tuple[Union[StrID, Expr], Expr]) -> None:
        super().__init__()
        self.elements = [(ID(n) if isinstance(n, str) else n, e) for n, e in elements]

    def _update_str(self) -> None:
        assert len(self.elements) > 0
        self._str = intern(
            "(" + ", ".join(f"{name} => {element}" for name, element in self.elements) + ")"
        )

    @property
    def precedence(self) -> Precedence:
        raise NotImplementedError


class Relation(BinExpr):
    @property
    def precedence(self) -> Precedence:
        return Precedence.RELATIONAL_OPERATOR


class Less(Relation):
    @property
    def symbol(self) -> str:
        return " < "


class LessEqual(Relation):
    @property
    def symbol(self) -> str:
        return " <= "


class Equal(Relation):
    @property
    def symbol(self) -> str:
        return " = "


class GreaterEqual(Relation):
    @property
    def symbol(self) -> str:
        return " >= "


class Greater(Relation):
    @property
    def symbol(self) -> str:
        return " > "


class NotEqual(Relation):
    @property
    def symbol(self) -> str:
        return " /= "


class In(Relation):
    @property
    def symbol(self) -> str:
        return " in "


class NotIn(Relation):
    @property
    def symbol(self) -> str:
        return " not in "


def If(condition_expressions: Sequence[Tuple[Expr, Expr]], else_expression: Expr = None) -> Expr:
    # pylint: disable=invalid-name
    if len(condition_expressions) == 1 and condition_expressions[0][0] == TRUE:
        return condition_expressions[0][1]
    return IfExpr(condition_expressions, else_expression)


class IfExpr(Expr):
    def __init__(
        self, condition_expressions: Sequence[Tuple[Expr, Expr]], else_expression: Expr = None
    ) -> None:
        super().__init__()
        self.condition_expressions = condition_expressions
        self.else_expression = else_expression

    def _update_str(self) -> None:
        condition_expressions = [(str(c), str(e)) for c, e in self.condition_expressions]
        else_expression = str(self.else_expression)

        expression = "".join(
            f"if {c} then {e}" if i == 0 else f" elsif {c} then {e}"
            for i, (c, e) in enumerate(condition_expressions)
        )
        if self.else_expression:
            expression += f" else {else_expression}"
        expression = " ".join(expression.split())

        if len(expression) > 100:
            expression = ""
            expression = "".join(
                f"if\n{indent(c, 4)}\n then\n{indent(e, 4)}"
                if i == 0
                else f"\n elsif\n{indent(c, 4)}\n then\n{indent(e, 4)}"
                for i, (c, e) in enumerate(condition_expressions)
            )
            if self.else_expression:
                expression += f"\n else\n{indent(else_expression, 4)}"

        self._str = intern(f"({expression})")

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


def Case(control_expression: Expr, case_expressions: Sequence[Tuple[Expr, Expr]]) -> Expr:
    # pylint: disable=invalid-name
    if len(case_expressions) == 1 and case_expressions[0][0] == Variable("others"):
        return case_expressions[0][1]
    return CaseExpr(control_expression, case_expressions)


class CaseExpr(Expr):
    def __init__(
        self, control_expression: Expr, case_expressions: Sequence[Tuple[Expr, Expr]]
    ) -> None:
        super().__init__()
        self.control_expression = control_expression
        self.case_expressions = case_expressions

    def _update_str(self) -> None:
        grouped_cases = [
            (" | ".join(str(c) for c, _ in choices), expr)
            for expr, choices in itertools.groupby(self.case_expressions, lambda x: x[1])
        ]
        cases = indent(
            ",".join(
                [f"\nwhen {choice} =>\n{indent(str(expr), 3)}" for choice, expr in grouped_cases]
            ),
            4,
        )
        self._str = intern(f"(case {self.control_expression} is{cases})")

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class QuantifiedExpr(Expr):
    def __init__(self, parameter_identifier: StrID, iterable: Expr, predicate: Expr) -> None:
        super().__init__()
        self.parameter_identifier = ID(parameter_identifier)
        self.iterable = iterable
        self.predicate = predicate

    def _update_str(self) -> None:
        self._str = intern(
            f"(for {self.quantifier} {self.parameter_identifier} {self.keyword} {self.iterable}"
            f" =>\n{indent(str(self.predicate), 4)})"
        )

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL

    @property
    @abstractmethod
    def quantifier(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def keyword(self) -> str:
        raise NotImplementedError


class ForAllOf(QuantifiedExpr):
    @property
    def quantifier(self) -> str:
        return "all"

    @property
    def keyword(self) -> str:
        return "of"


class ForAllIn(QuantifiedExpr):
    @property
    def quantifier(self) -> str:
        return "all"

    @property
    def keyword(self) -> str:
        return "in"


class ForSomeIn(QuantifiedExpr):
    @property
    def quantifier(self) -> str:
        return "some"

    @property
    def keyword(self) -> str:
        return "in"


class ValueRange(Expr):
    def __init__(self, lower: Expr, upper: Expr):
        super().__init__()
        self.lower = lower
        self.upper = upper

    def _update_str(self) -> None:
        self._str = intern(f"{self.lower} .. {self.upper}")

    @property
    def precedence(self) -> Precedence:
        raise NotImplementedError


class Conversion(Expr):
    def __init__(self, identifier: StrID, argument: Expr) -> None:
        super().__init__()
        self.identifier = ID(identifier)
        self.argument = argument

    def _update_str(self) -> None:
        self._str = intern(f"{self.identifier} ({self.argument})")

    @property
    def precedence(self) -> Precedence:
        return Precedence.LITERAL


class QualifiedExpr(Expr):
    def __init__(self, type_identifier: StrID, expression: Expr) -> None:
        super().__init__()
        self.type_identifier = ID(type_identifier)
        self.expression = expression

    def _update_str(self) -> None:
        operand = (
            str(self.expression)
            if isinstance(self.expression, (Aggregate, NamedAggregate))
            else f"({self.expression})"
        )
        self._str = intern(f"{self.type_identifier}'{operand}")

    @property
    def precedence(self) -> Precedence:
        raise NotImplementedError


class New(Expr):
    def __init__(self, expression: Expr) -> None:
        super().__init__()
        self.expression = expression

    def _update_str(self) -> None:
        self._str = intern(f"new {self.expression}")

    @property
    def precedence(self) -> Precedence:
        raise NotImplementedError


class Raise(Expr):
    def __init__(self, identifier: StrID, string: Expr = None) -> None:
        super().__init__()
        self.identifier = ID(identifier)
        self.string = string

    def _update_str(self) -> None:
        string = f" with {self.string}" if self.string else ""
        self._str = intern(f"raise {self.identifier}{string}")

    @property
    def precedence(self) -> Precedence:
        raise NotImplementedError


class Declaration(Base):
    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class ContextItem(Base):
    def __init__(self, identifier: StrID) -> None:
        self.identifier = ID(identifier)

    def __hash__(self) -> int:
        return hash(self.identifier)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class WithClause(ContextItem):
    def __str__(self) -> str:
        return f"with {self.identifier};"


class UsePackageClause(ContextItem, Declaration):
    def __str__(self) -> str:
        return f"use {self.identifier};"


class UseTypeClause(ContextItem, Declaration):
    def __str__(self) -> str:
        return f"use type {self.identifier};"


class Aspect(Base):
    def __str__(self) -> str:
        if self.definition:
            return f"{self.mark} =>\n{indent(self.definition, 2)}"
        return f"{self.mark}"

    @property
    @abstractmethod
    def mark(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def definition(self) -> str:
        raise NotImplementedError


class Precondition(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Pre"

    @property
    def definition(self) -> str:
        return str(self.expr)


class Postcondition(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Post"

    @property
    def definition(self) -> str:
        return str(self.expr)


class ContractCases(Aspect):
    def __init__(self, *cases: Tuple[Expr, Expr]) -> None:
        self.cases = cases

    @property
    def mark(self) -> str:
        return "Contract_Cases"

    @property
    def definition(self) -> str:
        cases = indent_next(",\n".join(f"{p} =>\n{indent(str(q), 3)}" for p, q in self.cases), 1)
        return f"({cases})"


class Depends(Aspect):
    def __init__(self, dependencies: Mapping[StrID, Sequence[StrID]]) -> None:
        self.dependencies = dependencies

    @property
    def mark(self) -> str:
        return "Depends"

    @property
    def definition(self) -> str:
        def input_values(values: Sequence[StrID]) -> str:
            if len(values) == 0:
                return "null"
            if len(values) == 1:
                return str(values[0])
            return "(" + ", ".join(str(p) for p in values) + ")"

        dependencies = indent_next(
            ", ".join(f"{o} => {input_values(i)}" for o, i in self.dependencies.items()), 1
        )
        return f"({dependencies})"


class DynamicPredicate(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Dynamic_Predicate"

    @property
    def definition(self) -> str:
        return str(self.expr)


class SizeAspect(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Size"

    @property
    def definition(self) -> str:
        return str(self.expr)


class InitialCondition(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Initial_Condition"

    @property
    def definition(self) -> str:
        return str(self.expr)


class DefaultInitialCondition(Aspect):
    def __init__(self, expr: Expr) -> None:
        self.expr = expr

    @property
    def mark(self) -> str:
        return "Default_Initial_Condition"

    @property
    def definition(self) -> str:
        return str(self.expr)


class SparkMode(Aspect):
    @property
    def mark(self) -> str:
        return "SPARK_Mode"

    @property
    def definition(self) -> str:
        return ""


class Ghost(Aspect):
    @property
    def mark(self) -> str:
        return "Ghost"

    @property
    def definition(self) -> str:
        return ""


class Import(Aspect):
    @property
    def mark(self) -> str:
        return "Import"

    @property
    def definition(self) -> str:
        return ""


class Annotate(Aspect):
    def __init__(self, *args: str) -> None:
        self.args = args

    @property
    def mark(self) -> str:
        return "Annotate"

    @property
    def definition(self) -> str:
        args = ", ".join(self.args)
        return f"({args})"


class FormalDeclaration(Base):
    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class FormalSubprogramDeclaration(FormalDeclaration):
    def __init__(self, specification: "SubprogramSpecification") -> None:
        self.specification = specification

    def __hash__(self) -> int:
        return hash(self.specification)

    def __str__(self) -> str:
        return f"with {self.specification};"


class FormalPackageDeclaration(FormalDeclaration):
    def __init__(
        self, identifier: StrID, generic_identifier: StrID, associations: Sequence[StrID] = None
    ) -> None:
        self.identifier = ID(identifier)
        self.generic_identifier = ID(generic_identifier)
        self.associations = list(map(str, associations or []))

    def __str__(self) -> str:
        associations = ", ".join(map(str, self.associations)) if self.associations else "<>"
        return f"with package {self.identifier} is new {self.generic_identifier} ({associations});"


class PackageDeclaration(Declaration):
    def __init__(
        self,
        identifier: StrID,
        declarations: List[Declaration] = None,
        private_declarations: List[Declaration] = None,
        formal_parameters: List[FormalDeclaration] = None,
        aspects: List[Aspect] = None,
    ) -> None:
        self.identifier = ID(identifier)
        self.declarations = declarations or []
        self.private_declarations = private_declarations or []
        self.formal_parameters = formal_parameters
        self.aspects = aspects or []

    def __str__(self) -> str:
        return (
            f"{generic_formal_part(self.formal_parameters)}"
            f"package {self.identifier}{aspect_specification(self.aspects)}\nis\n\n"
            f"{declarative_items(self.declarations)}"
            f"{declarative_items(self.private_declarations, private=True)}"
            f"end {self.identifier};\n"
        )


class PackageBody(Declaration):
    def __init__(
        self,
        identifier: StrID,
        declarations: List[Declaration] = None,
        statements: List["Statement"] = None,
        aspects: List[Aspect] = None,
    ) -> None:
        self.identifier = ID(identifier)
        self.declarations = declarations or []
        self.statements = statements or []
        self.aspects = aspects or []

    def __str__(self) -> str:
        if not self.declarations and not self.statements:
            return ""

        statements = (
            ("begin\n\n" + indent("\n".join(map(str, self.statements)), 3) + "\n\n")
            if self.statements
            else ""
        )

        return (
            f"package body {self.identifier}{aspect_specification(self.aspects)}\nis\n\n"
            f"{declarative_items(self.declarations)}{statements}end {self.identifier};\n"
        )


class GenericPackageInstantiation(Declaration):
    def __init__(
        self, identifier: StrID, generic_package: StrID, associations: Sequence[StrID] = None
    ) -> None:
        self.identifier = ID(identifier)
        self.generic_package = ID(generic_package)
        self.associations = list(map(ID, associations or []))

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __str__(self) -> str:
        associations = ", ".join(map(str, self.associations))
        if associations:
            associations = f" ({associations})"
        return f"package {self.identifier} is new {self.generic_package}{associations};"


class PackageRenamingDeclaration(Declaration):
    def __init__(self, identifier: StrID, package_identifier: StrID) -> None:
        self.identifier = ID(identifier)
        self.package_identifier = ID(package_identifier)

    def __str__(self) -> str:
        return f"package {self.identifier} renames {self.package_identifier};"


class ObjectDeclaration(Declaration):
    def __init__(
        self,
        identifiers: Sequence[StrID],
        type_identifier: Union[StrID, Expr],
        expression: Expr = None,
        constant: bool = False,
        aspects: Sequence[Aspect] = None,
    ) -> None:
        self.identifiers = list(map(ID, identifiers))
        self.type_identifier = (
            type_identifier if isinstance(type_identifier, Expr) else Variable(type_identifier)
        )
        self.expression = expression
        self.constant = constant
        self.aspects = aspects or []

    def __hash__(self) -> int:
        return hash(tuple(self.identifiers))

    def __str__(self) -> str:
        identifiers = ", ".join(map(str, self.identifiers))
        constant = "constant " if self.constant else ""
        expression = f" := {self.expression}" if self.expression else ""
        return (
            f"{identifiers} : {constant}{self.type_identifier}{expression}"
            f"{aspect_specification(self.aspects)};"
        )


class Discriminant(Base):
    def __init__(
        self, identifiers: Sequence[StrID], type_identifier: StrID, default: Expr = None
    ) -> None:
        self.identifiers = list(map(ID, identifiers))
        self.type_identifier = ID(type_identifier)
        self.default = default

    def __str__(self) -> str:
        identifiers = ", ".join(map(str, self.identifiers))
        default = f" := {self.default}" if self.default else ""
        return f"{identifiers} : {self.type_identifier}{default}"


class TypeDeclaration(Declaration, FormalDeclaration):
    def __init__(
        self,
        identifier: StrID,
        discriminants: Sequence[Discriminant] = None,
        aspects: Sequence[Aspect] = None,
    ) -> None:
        self.identifier = ID(identifier)
        self.discriminants = discriminants
        self.aspects = aspects or []

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __str__(self) -> str:
        return (
            f"type {self.identifier}{self.discriminant_part} is{self.type_definition}"
            f"{aspect_specification(self.aspects)};{self.extra_declaration}"
        )

    @property
    def discriminant_part(self) -> str:
        return " (" + "; ".join(map(str, self.discriminants)) + ")" if self.discriminants else ""

    @property
    @abstractmethod
    def type_definition(self) -> str:
        raise NotImplementedError

    @property
    def extra_declaration(self) -> str:
        return ""


class ModularType(TypeDeclaration):
    def __init__(self, identifier: StrID, modulus: Expr, aspects: Sequence[Aspect] = None) -> None:
        super().__init__(identifier, aspects=aspects or [])
        self.modulus = modulus

    @property
    def type_definition(self) -> str:
        return f" mod {self.modulus}"


class RangeType(TypeDeclaration):
    def __init__(
        self,
        identifier: StrID,
        first: Expr,
        last: Expr,
        aspects: Sequence[Aspect] = None,
    ) -> None:
        super().__init__(identifier, aspects=aspects or [])
        self.first = first
        self.last = last

    @property
    def type_definition(self) -> str:
        return f" range {self.first} .. {self.last}"


class EnumerationType(TypeDeclaration):
    def __init__(
        self,
        identifier: StrID,
        literals: Mapping[ID, Optional[Number]],
        size: Expr = None,
    ) -> None:
        super().__init__(identifier, aspects=([SizeAspect(size)] if size else []))
        self.literals = (
            OrderedDict(
                sorted(
                    literals.items(), key=lambda t: t[1].value if isinstance(t[1], Number) else 0
                )
            )
            if None not in literals.values()
            else literals
        )
        self.size = size

    @property
    def type_definition(self) -> str:
        literal_specification = ", ".join(map(str, self.literals.keys()))
        return f" ({literal_specification})"

    @property
    def extra_declaration(self) -> str:
        literal_representation = ", ".join(
            f"{k} => {v}" for k, v in self.literals.items() if v is not None
        )
        return (
            f"\nfor {self.identifier} use ({literal_representation});"
            if literal_representation
            else ""
        )


class Subtype(TypeDeclaration):
    def __init__(self, identifier: StrID, base_identifier: StrID) -> None:
        super().__init__(identifier)
        self.base_identifier = ID(base_identifier)

    def __str__(self) -> str:
        return "sub" + super().__str__()

    @property
    def type_definition(self) -> str:
        return f" {self.base_identifier}"


class RangeSubtype(Subtype):
    def __init__(self, identifier: StrID, base_identifier: StrID, first: Expr, last: Expr) -> None:
        super().__init__(identifier, base_identifier)
        self.first = first
        self.last = last

    @property
    def type_definition(self) -> str:
        return f" {self.base_identifier} range {self.first} .. {self.last}"


class DerivedType(TypeDeclaration):
    def __init__(self, identifier: StrID, type_identifier: StrID) -> None:
        super().__init__(identifier)
        self.type_identifier = ID(type_identifier)

    @property
    def type_definition(self) -> str:
        return f" new {self.type_identifier}"


class PrivateType(TypeDeclaration):
    @property
    def type_definition(self) -> str:
        return " private"


class DiscreteType(TypeDeclaration):
    @property
    def type_definition(self) -> str:
        return " (<>)"


class ArrayType(TypeDeclaration):
    def __init__(self, identifier: StrID, index_type: StrID, component_identifier: StrID) -> None:
        super().__init__(identifier)
        self.index_type = ID(index_type)
        self.component_identifier = ID(component_identifier)

    @property
    def type_definition(self) -> str:
        return f" array ({self.index_type}) of {self.component_identifier}"


class UnconstrainedArrayType(ArrayType):
    @property
    def type_definition(self) -> str:
        return f" array ({self.index_type} range <>) of {self.component_identifier}"


class AccessType(TypeDeclaration):
    def __init__(self, identifier: StrID, object_identifier: StrID) -> None:
        super().__init__(identifier)
        self.object_identifier = ID(object_identifier)

    @property
    def type_definition(self) -> str:
        return f" access {self.object_identifier}"


class Component(Base):
    def __init__(
        self, identifier: StrID, type_identifier: Union[StrID, Expr], default: Expr = None
    ) -> None:
        self.identifier = ID(identifier)
        self.type_identifier = (
            type_identifier if isinstance(type_identifier, Expr) else Variable(type_identifier)
        )
        self.default = default

    def __str__(self) -> str:
        default = f" := {self.default}" if self.default else ""
        return f"{self.identifier} : {self.type_identifier}{default};"


class NullComponent(Component):
    def __init__(self) -> None:
        super().__init__("null", "null")

    def __str__(self) -> str:
        return "null;"


class Variant(Base):
    def __init__(self, discrete_choices: Sequence[Expr], components: Sequence[Component]) -> None:
        self.discrete_choices = discrete_choices
        self.components = components

    def __str__(self) -> str:
        choices = " | ".join(map(str, self.discrete_choices))
        components = indent("\n".join(map(str, self.components)), 6)
        return f"   when {choices} =>\n{components}"


class VariantPart(Base):
    def __init__(self, discriminant_identifier: StrID, variants: Sequence[Variant]) -> None:
        self.discriminant_identifier = ID(discriminant_identifier)
        self.variants = variants

    def __str__(self) -> str:
        variants = "\n".join(map(str, self.variants))
        return f"case {self.discriminant_identifier} is\n{variants}\nend case;\n"


class RecordType(TypeDeclaration):
    def __init__(
        self,
        identifier: StrID,
        components: List[Component],
        discriminants: Sequence[Discriminant] = None,
        variant_part: VariantPart = None,
        aspects: Sequence[Aspect] = None,
    ) -> None:
        super().__init__(identifier, discriminants, aspects)
        self.components = components
        self.discriminants = discriminants or []
        self.variant_part = variant_part

    @property
    def type_definition(self) -> str:
        components = (
            (indent("\n".join(map(str, self.components)), 6) + "\n") if self.components else ""
        )
        variant_part = indent(str(self.variant_part), 6) if self.variant_part else ""
        return f"\n" f"   record\n" f"{components}" f"{variant_part}" f"   end record"


class Statement(Base):
    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class NullStatement(Statement):
    def __init__(self) -> None:
        pass

    def __str__(self) -> str:
        return "null;"


class Assignment(Statement):
    def __init__(self, identifier: Union[StrID, Expr], expression: Expr) -> None:
        self.identifier = identifier if isinstance(identifier, Expr) else Variable(identifier)
        self.expression = expression

    def __str__(self) -> str:
        return f"{self.identifier} := {self.expression};"


class CallStatement(Statement):
    def __init__(self, identifier: StrID, arguments: Sequence[Expr] = None) -> None:
        self.identifier = ID(identifier)
        self.arguments = arguments or []

    def __str__(self) -> str:
        arguments = ", ".join(map(str, self.arguments))
        arguments = f" ({arguments})" if arguments else ""
        return f"{self.identifier}{arguments};"


class PragmaStatement(Statement):
    def __init__(self, identifier: StrID, parameters: Sequence[Expr]) -> None:
        self.identifier = ID(identifier)
        self.pragma_parameters = parameters

    def __str__(self) -> str:
        parameters = ""
        if self.pragma_parameters:
            parameters = ", ".join(map(str, self.pragma_parameters))
            parameters = " (" + indent_next(str(parameters), len(str(self.identifier)) + 9) + ")"
        return f"pragma {self.identifier}{parameters};"


class ReturnStatement(Statement):
    def __init__(self, expression: Expr = None) -> None:
        self.expression = expression

    def __str__(self) -> str:
        if not self.expression:
            return "return;"
        if isinstance(self.expression, CaseExpr):
            return "return (" + indent_next(str(self.expression), 8) + ");"
        return "return " + indent_next(str(self.expression), 7) + ";"


class ExitStatement(Statement):
    def __init__(self, expression: Expr = None) -> None:
        self.expression = expression

    def __str__(self) -> str:
        if not self.expression:
            return "exit;"
        if isinstance(self.expression, CaseExpr):
            return "exit when (" + indent_next(str(self.expression), 8) + ");"
        return "exit when " + indent_next(str(self.expression), 7) + ";"


class IfStatement(Statement):
    def __init__(
        self,
        condition_statements: Sequence[Tuple[Expr, Sequence[Statement]]],
        else_statements: Sequence[Statement] = None,
    ) -> None:
        assert condition_statements or else_statements
        self.condition_statements = condition_statements
        self.else_statements = else_statements

    def __str__(self) -> str:
        result = ""

        for condition, statements in self.condition_statements:
            c = (
                f" {condition} "
                if str(condition).count("\n") == 0
                else f"\n{indent(str(condition), 2)}\n"
            )
            if not result:
                result = f"if{c}then\n"
            else:
                result += f"elsif{c}then\n"
            for statement in statements:
                result += indent(f"{statement}\n", 3)

        if self.else_statements:
            else_statements = "\n".join(map(str, self.else_statements))

            if not self.condition_statements:
                return else_statements

            result += f"else\n{indent(else_statements, 3)}\n"

        result += "end if;"

        return result


class CaseStatement(Statement):
    def __init__(
        self,
        control_expression: Expr,
        case_statements: Sequence[Tuple[Expr, Sequence[Statement]]],
        case_grouping: bool = True,
    ) -> None:
        self.control_expression = control_expression
        self.case_statements = case_statements
        self.case_grouping = case_grouping

    def __str__(self) -> str:
        grouped_cases = (
            [
                (" | ".join(str(c) for c, _ in choices), statements)
                for statements, choices in itertools.groupby(self.case_statements, lambda x: x[1])
            ]
            if self.case_grouping
            else [(str(case), statements) for case, statements in self.case_statements]
        )
        cases = "".join(
            [
                "\nwhen {} =>\n{}".format(choice, indent("\n".join(str(s) for s in statements), 3))
                for choice, statements in grouped_cases
            ]
        )

        return f"case {self.control_expression} is{indent(cases, 3)}\nend case;"


class While(Statement):
    def __init__(self, condition: Expr, statements: Sequence[Statement]) -> None:
        assert len(statements) > 0
        self.condition = condition
        self.statements = statements

    def __str__(self) -> str:
        statements = indent("\n".join(str(s) for s in self.statements), 3)
        return f"while {self.condition} loop\n{statements}\nend loop;"


class ForOf(Statement):
    def __init__(
        self,
        identifier: StrID,
        iterator: Expr,
        statements: Sequence[Statement],
        reverse: bool = False,
    ) -> None:
        assert len(statements) > 0
        self.identifier = identifier
        self.iterator = iterator
        self.statements = statements
        self.reverse = reverse

    def __str__(self) -> str:
        statements = indent("\n".join(str(s) for s in self.statements), 3)
        reverse = "reverse " if self.reverse else ""
        return f"for {self.identifier} of {reverse}{self.iterator} loop\n{statements}\nend loop;"


class Declare(Statement):
    def __init__(
        self, declarations: Sequence[Declaration], statements: Sequence[Statement]
    ) -> None:
        self.declarations = declarations
        self.statements = statements

    def __str__(self) -> str:
        declarations = indent("\n".join(str(s) for s in self.declarations), 3)
        statements = indent("\n".join(str(s) for s in self.statements), 3)
        return f"declare\n{declarations}\nbegin\n{statements}\nend;"


class Parameter(Base):
    def __init__(
        self, identifiers: Sequence[StrID], type_identifier: StrID, default: Expr = None
    ) -> None:
        self.identifiers = list(map(ID, identifiers))
        self.type_identifier = ID(type_identifier)
        self.default = default

    def __str__(self) -> str:
        identifiers = ", ".join(map(str, self.identifiers))
        default = f" := {self.default}" if self.default else ""
        return f"{identifiers} : {self.mode}{self.type_identifier}{default}"

    @property
    def mode(self) -> str:
        return ""


class OutParameter(Parameter):
    @property
    def mode(self) -> str:
        return "out "


class InOutParameter(Parameter):
    @property
    def mode(self) -> str:
        return "in out "


class AccessParameter(Parameter):
    def __init__(
        self,
        identifiers: Sequence[StrID],
        type_identifier: StrID,
        default: Expr = None,
        constant: bool = False,
    ) -> None:
        super().__init__(identifiers, type_identifier, default)
        self.constant = constant

    @property
    def mode(self) -> str:
        return "access constant " if self.constant else "access "


class SubprogramSpecification(Base):
    def __init__(self, identifier: StrID, parameters: Sequence[Parameter] = None) -> None:
        self.identifier = ID(identifier)
        self.parameters = parameters or []

    def _parameters(self) -> str:
        return (" (" + "; ".join(map(str, self.parameters)) + ")") if self.parameters else ""

    def __hash__(self) -> int:
        return hash(self.identifier)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class ProcedureSpecification(SubprogramSpecification):
    def __str__(self) -> str:
        return f"procedure {self.identifier}{self._parameters()}"


class FunctionSpecification(SubprogramSpecification):
    def __init__(
        self, identifier: StrID, return_type: StrID, parameters: Sequence[Parameter] = None
    ) -> None:
        super().__init__(identifier, parameters)
        self.return_type = ID(return_type)

    def __str__(self) -> str:
        return f"function {self.identifier}{self._parameters()} return {self.return_type}"


class Subprogram(Declaration):
    def __init__(self, specification: SubprogramSpecification) -> None:
        self.specification = specification

    def __hash__(self) -> int:
        return hash(self.specification)

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class SubprogramDeclaration(Subprogram):
    def __init__(
        self,
        specification: SubprogramSpecification,
        aspects: Sequence[Aspect] = None,
        formal_parameters: Sequence[FormalDeclaration] = None,
    ) -> None:
        super().__init__(specification)
        self.aspects = aspects or []
        self.formal_parameters = formal_parameters

    def __str__(self) -> str:
        return (
            f"{generic_formal_part(self.formal_parameters)}"
            f"{self.specification}{aspect_specification(self.aspects)};"
        )


class SubprogramBody(Subprogram):
    def __init__(
        self,
        specification: SubprogramSpecification,
        declarations: Sequence[Declaration],
        statements: Sequence[Statement],
        aspects: Sequence[Aspect] = None,
    ) -> None:
        super().__init__(specification)
        self.declarations = declarations or []
        self.statements = statements or []
        self.aspects = aspects or []

    def _declarations(self) -> str:
        return "".join(indent(f"{declaration}\n", 3) for declaration in self.declarations)

    def _statements(self) -> str:
        return "\n".join(indent(str(s), 3) for s in self.statements)

    def __str__(self) -> str:
        aspects = f"{aspect_specification(self.aspects)}\n" if self.aspects else " "
        return (
            f"{self.specification}{aspects}is\n"
            f"{self._declarations()}"
            f"begin\n"
            f"{self._statements()}\n"
            f"end {self.specification.identifier};"
        )


class ExpressionFunctionDeclaration(Subprogram):
    def __init__(
        self,
        specification: FunctionSpecification,
        expression: Expr,
        aspects: List[Aspect] = None,
    ) -> None:
        super().__init__(specification)
        self.expression = expression
        self.aspects = aspects or []

    def __str__(self) -> str:
        aspects = f"\n{aspect_specification(self.aspects)}" if self.aspects else ""
        return f"{self.specification} is\n  ({indent_next(str(self.expression), 3)}){aspects};"


class GenericProcedureInstantiation(Subprogram):
    def __init__(
        self,
        identifier: StrID,
        specification: ProcedureSpecification,
        associations: List[StrID] = None,
    ) -> None:
        super().__init__(specification)
        self.identifier = ID(identifier)
        self.parameters = specification.parameters
        self.associations = list(map(ID, associations or []))

    def __str__(self) -> str:
        associations = ", ".join(map(str, self.associations))
        if associations:
            associations = f" ({associations})"
        return f"procedure {self.identifier} is new {self.specification.identifier}{associations};"


class GenericFunctionInstantiation(Subprogram):
    def __init__(
        self,
        identifier: StrID,
        specification: FunctionSpecification,
        associations: List[StrID] = None,
    ) -> None:
        super().__init__(specification)
        self.identifier = ID(identifier)
        self.parameters = specification.parameters
        self.associations = list(map(ID, associations or []))

    def __str__(self) -> str:
        associations = ", ".join(map(str, self.associations))
        if associations:
            associations = f" ({associations})"
        return f"function {self.identifier} is new {self.specification.identifier}{associations};"


class SubprogramRenamingDeclaration(Subprogram):
    def __init__(
        self, specification: SubprogramSpecification, subprogram_identifier: StrID
    ) -> None:
        super().__init__(specification)
        self.subprogram_identifier = ID(subprogram_identifier)

    def __str__(self) -> str:
        return f"{self.specification} renames {self.subprogram_identifier};"


class Pragma(Declaration, ContextItem):
    def __init__(self, identifier: StrID, parameters: List[Expr] = None) -> None:
        super().__init__(identifier)
        self.pragma_parameters = parameters or []

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return False
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.identifier)

    def __str__(self) -> str:
        parameters = ""
        if self.pragma_parameters:
            parameters = ", ".join(map(str, self.pragma_parameters))
            parameters = f" ({parameters})"
        return f"pragma {self.identifier}{parameters};"


class Unit(Base):
    @abstractmethod
    def __iadd__(self, other: object) -> "Unit":
        raise NotImplementedError

    @property
    @abstractmethod
    def ads(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def adb(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError


@invariant(lambda self: self.declaration.identifier == self.body.identifier)
class PackageUnit(Unit):
    def __init__(
        self,
        declaration_context: List[ContextItem],
        declaration: PackageDeclaration,
        body_context: List[ContextItem],
        body: PackageBody,
    ) -> None:
        self.declaration_context = declaration_context
        self.declaration = declaration
        self.body_context = body_context
        self.body = body

    def __iadd__(self, other: object) -> "PackageUnit":
        if isinstance(other, (UnitPart, SubprogramUnitPart)):
            self.declaration.declarations = self.declaration.declarations + list(
                other.specification
            )
            self.declaration.private_declarations = self.declaration.private_declarations + list(
                other.private
            )
            self.body.declarations = self.body.declarations + list(other.body)
            self.body.statements += list(other.statements)
            return self
        return NotImplemented

    @property
    def ads(self) -> str:
        return f"{context_clause(self.declaration_context)}{self.declaration}"

    @property
    def adb(self) -> str:
        return f"{context_clause(self.body_context)}{self.body}" if str(self.body) else ""

    @property
    def name(self) -> str:
        return file_name(str(self.declaration.identifier))


class InstantiationUnit(Unit):
    def __init__(
        self, context: List[ContextItem], declaration: GenericPackageInstantiation
    ) -> None:
        self.context = context
        self.declaration = declaration

    def __iadd__(self, other: object) -> Unit:
        return NotImplemented

    @property
    def ads(self) -> str:
        return f"{context_clause(self.context)}{self.declaration}\n"

    @property
    def adb(self) -> str:
        return ""

    @property
    def name(self) -> str:
        return file_name(str(self.declaration.identifier))


@dataclass
class UnitPart:
    specification: List[Declaration] = dataclass_field(default_factory=list)
    body: List[Declaration] = dataclass_field(default_factory=list)
    private: List[Declaration] = dataclass_field(default_factory=list)
    statements: List[Statement] = dataclass_field(default_factory=list)

    def __iadd__(self, other: object) -> "UnitPart":
        if isinstance(other, UnitPart):
            self.specification += other.specification
            self.body += other.body
            self.private += other.private
            self.statements += other.statements
            return self
        return NotImplemented


@dataclass
class SubprogramUnitPart:
    specification: List[Subprogram] = dataclass_field(default_factory=list)
    body: List[Subprogram] = dataclass_field(default_factory=list)
    private: List[Subprogram] = dataclass_field(default_factory=list)
    statements: List[Statement] = dataclass_field(default_factory=list)


def generic_formal_part(parameters: Sequence[FormalDeclaration] = None) -> str:
    if parameters is None:
        return ""
    return (
        "generic" + ("".join(f"\n   {p}" for p in unique(parameters)) if parameters else "") + "\n"
    )


def declarative_items(declarations: List[Declaration], private: bool = False) -> str:
    result = "\n\n".join(indent(str(d), 3) for d in unique(declarations) if str(d))
    if result:
        result = f"private\n\n{result}" if private else result
        result += "\n\n"
    return result


def aspect_specification(aspects: Sequence[Aspect]) -> str:
    if not aspects:
        return ""
    return " with\n" + ",\n".join(indent(str(aspect), 2) for aspect in aspects)


def context_clause(context: Sequence[ContextItem]) -> str:
    return ("\n".join(map(str, unique(context))) + "\n\n") if context else ""



import logging
from pathlib import Path
from typing import Callable, Dict, Iterable, Iterator, Union

from rflx.error import RecordFluxError
from rflx.identifier import ID
from rflx.model import Model
from rflx.pyrflx.typevalue import MessageValue, RefinementValue
from rflx.specification import Parser

from . import PyRFLXError
from .package import Package

log = logging.getLogger(__name__)


class PyRFLX:
    def __init__(
        self,
        model: Model,
        skip_message_verification: bool = False,
    ) -> None:
        self.__packages: Dict[str, Package] = {}
        messages: Dict[ID, MessageValue] = {}

        for m in model.messages:
            p = str(m.package)
            if p not in self.__packages:
                self.__packages[p] = Package(p)
            message = MessageValue(m, skip_verification=skip_message_verification)
            messages[m.identifier] = message
            self.__packages[p][str(m.name)] = message

        for r in model.refinements:
            messages[r.pdu.identifier].add_refinement(
                RefinementValue(r, messages[r.sdu.identifier])
            )

    @classmethod
    def from_specs(
        cls,
        files: Iterable[Union[str, Path]],
        skip_model_verification: bool = False,
        skip_message_verification: bool = False,
    ) -> "PyRFLX":
        paths = list(map(Path, files))
        for p in paths:
            if not p.is_file():
                raise FileNotFoundError(f'file not found: "{p}"')
        parser = Parser(skip_model_verification)
        parser.parse(*paths)
        model = parser.create_model()
        return cls(model, skip_message_verification)

    def set_checksum_functions(self, functions: Dict[str, Dict[str, Callable]]) -> None:
        for identifier_str, checksum_field_function in functions.items():
            try:
                message_identifier = ID(identifier_str)
            except RecordFluxError as e:
                raise PyRFLXError(f'"{identifier_str}" is not a valid identifier: {e}') from e

            if len(message_identifier.parts) < 2:
                raise PyRFLXError(f'"{identifier_str}" is not a valid identifier')

            if str(message_identifier.parent) in self.__packages.keys():
                package = self[str(message_identifier.parent)]
                package.set_checksum_functions(
                    {str(message_identifier.name): checksum_field_function}
                )

    def __getitem__(self, key: str) -> Package:
        return self.__packages[key]

    def __iter__(self) -> Iterator[Package]:
        return self.__packages.values().__iter__()



if __debug__:

    # pylint: disable=unused-import
    from icontract import DBC, DBCMeta, ViolationError, ensure, invariant, require, snapshot

else:

    import abc
    import reprlib
    from typing import Any, Callable, Optional, Union

    class DBCMeta(abc.ABCMeta):  # type: ignore
        pass

    class DBC(metaclass=DBCMeta):  # type: ignore
        pass

    class ViolationError(AssertionError):  # type: ignore
        pass

    # pylint: disable=unused-argument

    def require(  # type: ignore
        condition: Callable[..., Any],
        description: Optional[str] = None,
        a_repr: reprlib.Repr = None,
        enabled: bool = __debug__,
        error: Optional[Union[Callable[..., Exception], type]] = None,
    ) -> Callable:
        return lambda x: x

    def snapshot(  # type: ignore
        capture: Callable[..., Any], name: Optional[str] = None, enabled: bool = __debug__
    ) -> Callable:
        return lambda x: x

    def ensure(  # type: ignore
        condition: Callable[..., Any],
        description: Optional[str] = None,
        a_repr: reprlib.Repr = None,
        enabled: bool = __debug__,
        error: Optional[Union[Callable[..., Exception], type]] = None,
    ) -> Callable:
        return lambda x: x

    def invariant(  # type: ignore
        condition: Callable[..., Any],
        description: Optional[str] = None,
        a_repr: reprlib.Repr = None,
        enabled: bool = __debug__,
        error: Optional[Union[Callable[..., Exception], type]] = None,
    ) -> Callable:
        return lambda x: x



from rflx.error import RecordFluxError, Severity, Subsystem


class PyRFLXError(RecordFluxError):
    def __init__(self, message: str) -> None:
        super().__init__()
        self.append(message, Subsystem.PYRFLX, Severity.ERROR)



import re
from typing import Optional, Sequence, TypeVar, Union

from rflx.error import Location, RecordFluxError, Severity, Subsystem

Self = TypeVar("Self", bound="ID")


class ID:
    def __init__(
        self, identifier: Union[str, Sequence[str], "ID"], location: Location = None
    ) -> None:
        self._parts: Sequence[str]
        self.location = location

        if isinstance(identifier, str):
            self._parts = re.split(r"\.|::", identifier)
        elif isinstance(identifier, list):
            self._parts = identifier
        elif isinstance(identifier, ID):
            self._parts = list(identifier.parts)
            self.location = location or identifier.location
        else:
            assert False, f'unexpected identifier type "{type(identifier).__name__}"'

        error = RecordFluxError()
        if not self._parts:
            error.append("empty identifier", Subsystem.ID, Severity.ERROR, location)
        elif "" in self._parts:
            error.append(
                f'empty part in identifier "{self}"', Subsystem.ID, Severity.ERROR, location
            )
        else:
            for c in [" ", ".", ":"]:
                if any(c in part for part in self._parts):
                    error.append(
                        f'"{c}" in identifier parts of "{self}"',
                        Subsystem.ID,
                        Severity.ERROR,
                        location,
                    )
        error.propagate()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            if len(self.parts) != len(other.parts):
                return False
            return all(s.lower() == o.lower() for s, o in zip(self.parts, other.parts))
        return NotImplemented

    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return str(self) < str(other)
        return NotImplemented

    def __hash__(self) -> int:
        return hash(tuple(map(str.lower, self.parts)))

    def __repr__(self) -> str:
        return f'ID("{self}")'

    def __str__(self) -> str:
        return self._separator.join(self.parts)

    def __add__(self: Self, other: object) -> Self:
        if isinstance(other, (str, ID)):
            return self.__class__(f"{self}{other}", self.__location(other))
        return NotImplemented

    def __radd__(self: Self, other: object) -> Self:
        if isinstance(other, (str, ID)):
            return self.__class__(f"{other}{self}", self.__location(other))
        return NotImplemented

    def __mul__(self: Self, other: object) -> Self:
        if isinstance(other, (str, ID)):
            if str(other) == "":
                return self.__class__(self, self.__location(other))
            return self.__class__(f"{self}{self._separator}{other}", self.__location(other))
        return NotImplemented

    def __rmul__(self: Self, other: object) -> Self:
        if isinstance(other, (str, ID)):
            if str(other) == "":
                return self.__class__(self, self.__location(other))
            return self.__class__(f"{other}{self._separator}{self}", self.__location(other))
        return NotImplemented

    def __location(self, other: object) -> Optional[Location]:
        if isinstance(other, str):
            return self.location
        if isinstance(other, ID):
            if self.location is None and other.location is None:
                return None
            if self.location is None:
                return other.location
            return self.location
        raise NotImplementedError

    @property
    def parts(self) -> Sequence[str]:
        return self._parts

    @property
    def name(self: Self) -> Self:
        return self.__class__(self._parts[-1])

    @property
    def parent(self: Self) -> Self:
        return self.__class__(self._parts[:-1])

    @property
    def flat(self: Self) -> str:
        return "_".join(self._parts)

    @property
    def _separator(self) -> str:
        return "::"


StrID = Union[str, ID]



# pylint: disable=too-many-lines
import typing as ty
from abc import abstractmethod
from dataclasses import dataclass

from rflx.common import Base
from rflx.const import BUILTINS_PACKAGE
from rflx.expression import (
    FALSE,
    TRUE,
    UNDEFINED,
    Add,
    And,
    Attribute,
    Expr,
    First,
    Last,
    Name,
    Number,
    Size,
    Sub,
    ValidChecksum,
    ValueRange,
    Variable,
)
from rflx.identifier import ID
from rflx.model import (
    FINAL,
    INITIAL,
    Composite,
    Enumeration,
    Field,
    Integer,
    Link,
    Message,
    Opaque,
    Refinement,
    Scalar,
    Sequence,
    Type,
)
from rflx.pyrflx.bitstring import Bitstring
from rflx.pyrflx.error import PyRFLXError, Severity, Subsystem


class TypeValue(Base):

    _value: ty.Any = None

    def __init__(self, vtype: Type) -> None:
        self._type = vtype

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._value == other._value and self._type == other._type
        return NotImplemented

    def equal_type(self, other: Type) -> bool:
        return isinstance(self._type, type(other))

    @property
    def name(self) -> str:
        return self._type.name

    @property
    def identifier(self) -> ID:
        return self._type.identifier

    @property
    def package(self) -> ID:
        return self._type.package

    @property
    def initialized(self) -> bool:
        return self._value is not None

    def _raise_initialized(self) -> None:
        if not self.initialized:
            raise PyRFLXError(f"value {self.identifier} not initialized")

    def clear(self) -> None:
        self._value = None

    @abstractmethod
    def assign(self, value: ty.Any, check: bool = True) -> None:
        raise NotImplementedError

    @abstractmethod
    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        raise NotImplementedError

    @property
    @abstractmethod
    def bitstring(self) -> Bitstring:
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self) -> Expr:
        raise NotImplementedError

    @property
    @abstractmethod
    def value(self) -> ty.Any:
        raise NotImplementedError

    @property
    @abstractmethod
    def accepted_type(self) -> type:
        raise NotImplementedError

    def clone(self) -> "TypeValue":
        return self.__class__(self._type)

    @classmethod
    def construct(
        cls, vtype: Type, imported: bool = False, refinements: ty.Sequence["RefinementValue"] = None
    ) -> "TypeValue":
        if isinstance(vtype, Integer):
            return IntegerValue(vtype)
        if isinstance(vtype, Enumeration):
            return EnumValue(vtype, imported)
        if isinstance(vtype, Opaque):
            return OpaqueValue(vtype)
        if isinstance(vtype, Sequence):
            return SequenceValue(vtype)
        if isinstance(vtype, Message):
            return MessageValue(vtype, refinements)
        raise PyRFLXError("cannot construct unknown type: " + type(vtype).__name__)


class ScalarValue(TypeValue):

    _type: Scalar

    def __init__(self, vtype: Scalar) -> None:
        super().__init__(vtype)

    @property
    @abstractmethod
    def expr(self) -> Expr:
        return NotImplemented

    @property
    def size(self) -> Number:
        return self._type.size


class IntegerValue(ScalarValue):

    _value: int
    _type: Integer

    def __init__(self, vtype: Integer) -> None:
        super().__init__(vtype)

    @property
    def _first(self) -> int:
        return self._type.first.value

    @property
    def _last(self) -> int:
        return self._type.last.value

    def assign(self, value: int, check: bool = True) -> None:
        if check and (
            And(*self._type.constraints("__VALUE__", check))
            .substituted(
                mapping={Variable("__VALUE__"): Number(value), Size("__VALUE__"): self._type.size}
            )
            .simplified()
            != TRUE
        ):
            raise PyRFLXError(f"value {value} not in type range {self._first} .. {self._last}")
        self._value = value

    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        if isinstance(value, bytes):
            value = Bitstring.from_bytes(value)
        self.assign(int(value), check)

    @property
    def expr(self) -> Number:
        self._raise_initialized()
        return Number(self._value)

    @property
    def value(self) -> int:
        self._raise_initialized()
        return self._value

    @property
    def bitstring(self) -> Bitstring:
        self._raise_initialized()
        return Bitstring(format(self._value, f"0{self.size}b"))

    @property
    def accepted_type(self) -> type:
        return int


class EnumValue(ScalarValue):

    _value: ty.Tuple[str, Number]
    _type: Enumeration

    def __init__(self, vtype: Enumeration, imported: bool = False) -> None:
        super().__init__(vtype)
        self.__imported = imported
        self.__builtin = self._type.package == BUILTINS_PACKAGE
        self.__literals: ty.Dict[Name, Expr] = {}

        for k, v in self._type.literals.items():
            if self.__builtin or not self.__imported:
                self.__literals[Variable(k)] = v
            if not self.__builtin:
                self.__literals[Variable(self._type.package * k)] = v

    def assign(self, value: str, check: bool = True) -> None:
        prefixed_value = (
            ID(value)
            if value.startswith(str(self._type.package)) or not self.__imported or self.__builtin
            else self._type.package * value
        )
        if Variable(prefixed_value) not in self.literals:
            raise PyRFLXError(f"{value} is not a valid enum value")
        r = (
            (
                And(*self._type.constraints("__VALUE__", check, not self.__imported))
                .substituted(
                    mapping={
                        **self.literals,
                        **{Variable("__VALUE__"): self._type.literals[prefixed_value.name]},
                        **{Size("__VALUE__"): self._type.size},
                    }
                )
                .simplified()
            )
            if check
            else TRUE
        )
        assert r == TRUE
        self._value = (
            str(prefixed_value)
            if self.__imported and not self.__builtin
            else str(prefixed_value.name),
            self._type.literals[prefixed_value.name],
        )

    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        if isinstance(value, bytes):
            value = Bitstring.from_bytes(value)
        value_as_number = Number(int(value))
        if value_as_number not in self.literals.values():
            if self._type.always_valid:
                self._value = f"RFLX_UNKNOWN_{self.name.upper()}", value_as_number
            else:
                raise PyRFLXError(f"Number {value_as_number.value} is not a valid enum value")
        else:
            for k, v in self.literals.items():
                if v == value_as_number:
                    assert isinstance(k, Variable)
                    assert isinstance(v, Number)
                    self._value = (
                        str(k.identifier) if self.__imported else str(k.identifier.name),
                        v,
                    )

    def clone(self) -> "TypeValue":
        return self.__class__(self._type, self.__imported)

    @property
    def numeric_value(self) -> Number:
        self._raise_initialized()
        return self._value[1]

    @property
    def value(self) -> str:
        self._raise_initialized()
        return self._value[0]

    @property
    def expr(self) -> Variable:
        self._raise_initialized()
        return Variable(self._value[0])

    @property
    def bitstring(self) -> Bitstring:
        self._raise_initialized()
        return Bitstring(format(self._value[1].value, f"0{self.size}b"))

    @property
    def accepted_type(self) -> type:
        return str

    @property
    def literals(self) -> ty.Mapping[Name, Expr]:
        return self.__literals


class CompositeValue(TypeValue):
    def __init__(self, vtype: Composite) -> None:
        self._expected_size: ty.Optional[Expr] = None
        super().__init__(vtype)

    def set_expected_size(self, expected_size: Expr) -> None:
        self._expected_size = expected_size

    def _check_size_of_assigned_value(
        self, value: ty.Union[bytes, Bitstring, ty.List[TypeValue]]
    ) -> None:
        if isinstance(value, bytes):
            size_of_value = len(value) * 8
        elif isinstance(value, Bitstring):
            size_of_value = len(value)
        else:
            bits = [element.bitstring for element in value]
            size_of_value = len(Bitstring.join(bits))

        if (
            self._expected_size is not None
            and isinstance(self._expected_size, Number)
            and size_of_value != self._expected_size.value
        ):
            raise PyRFLXError(
                f"invalid data size: input size is {len(value) * 8} "
                f"while expected input size is {self._expected_size.value}"
            )

    @property
    @abstractmethod
    def value(self) -> ty.Any:
        raise NotImplementedError


class OpaqueValue(CompositeValue):

    _value: ty.Optional[bytes]
    _nested_message: ty.Optional["MessageValue"] = None

    def __init__(self, vtype: Opaque) -> None:
        super().__init__(vtype)
        self._refinement_message: ty.Optional["MessageValue"] = None

    def assign(self, value: bytes, check: bool = True) -> None:
        self.parse(value, check)

    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        if check:
            self._check_size_of_assigned_value(value)
        if self._refinement_message is not None:
            nested_msg = self._refinement_message.clone()
            try:
                nested_msg.parse(value, check)
            except PyRFLXError as e:
                e.appendleft(
                    f"Error while parsing nested message {self._refinement_message.identifier}",
                    Subsystem.PYRFLX,
                    Severity.ERROR,
                )
                raise e
            assert nested_msg.valid_message
            self._nested_message = nested_msg
            self._value = nested_msg.bytestring
        else:
            self._value = bytes(value)

    def set_refinement(self, model_of_refinement_msg: "MessageValue") -> None:
        self._refinement_message = model_of_refinement_msg

    @property
    def size(self) -> Expr:
        if self._value is None:
            return self._expected_size if self._expected_size is not None else UNDEFINED
        return Number(len(self._value) * 8)

    @property
    def nested_message(self) -> ty.Optional["MessageValue"]:
        return self._nested_message

    @property
    def value(self) -> bytes:
        self._raise_initialized()
        assert self._value is not None
        return self._value

    @property
    def bitstring(self) -> Bitstring:
        self._raise_initialized()
        assert self._value is not None
        size = self.size
        assert isinstance(size, Number)
        if size.value == 0:
            return Bitstring("")
        return Bitstring(format(int.from_bytes(self._value, "big"), f"0{size}b"))

    @property
    def accepted_type(self) -> type:
        return bytes


class SequenceValue(CompositeValue):

    _value: ty.List[TypeValue]

    def __init__(self, vtype: Sequence) -> None:
        super().__init__(vtype)
        self._element_type = vtype.element_type
        self._is_message_sequence = isinstance(self._element_type, Message)
        self._value = []

    def assign(self, value: ty.List[TypeValue], check: bool = True) -> None:
        if check:
            self._check_size_of_assigned_value(value)
        for v in value:
            if self._is_message_sequence:
                if isinstance(v, MessageValue):
                    assert isinstance(self._element_type, Message)
                    if not v.equal_type(self._element_type):
                        raise PyRFLXError(
                            f'cannot assign "{v.name}" to an sequence of '
                            f'"{self._element_type.name}"'
                        )
                    if not v.valid_message:
                        raise PyRFLXError(
                            f'cannot assign message "{v.name}" to sequence of messages: '
                            f"all messages must be valid"
                        )
                else:
                    raise PyRFLXError(
                        f"cannot assign {type(v).__name__} to an sequence of "
                        f"{type(self._element_type).__name__}"
                    )
            else:
                if isinstance(v, MessageValue) or not v.equal_type(self._element_type):
                    raise PyRFLXError(
                        f"cannot assign {type(v).__name__} to an sequence of "
                        f"{type(self._element_type).__name__}"
                    )

        self._value = value

    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        self._check_size_of_assigned_value(value)
        if isinstance(value, bytes):
            value = Bitstring.from_bytes(value)
        if self._is_message_sequence:

            while len(value) != 0:
                nested_message = TypeValue.construct(self._element_type)
                assert isinstance(nested_message, MessageValue)
                try:
                    nested_message.parse(value, check)
                except PyRFLXError as e:
                    e.appendleft(
                        f"cannot parse nested messages in sequence of type "
                        f"{self._element_type.full_name}",
                        Subsystem.PYRFLX,
                        Severity.ERROR,
                    )
                    raise e
                assert nested_message.valid_message
                self._value.append(nested_message)
                value = value[len(nested_message.bitstring) :]

        elif isinstance(self._element_type, Scalar):
            value_str = str(value)
            type_size = self._element_type.size
            type_size_int = type_size.value
            new_value = []

            while len(value_str) != 0:
                nested_value = TypeValue.construct(
                    self._element_type, imported=self._element_type.package != self._type.package
                )
                nested_value.parse(Bitstring(value_str[:type_size_int]), check)
                new_value.append(nested_value)
                value_str = value_str[type_size_int:]

            self._value = new_value
        else:
            raise PyRFLXError(
                f"sequences of {self._element_type.identifier} currently not supported"
            )

    @property
    def size(self) -> Expr:
        if not self._value:
            return self._expected_size if self._expected_size is not None else UNDEFINED
        return Number(len(self.bitstring))

    @property
    def value(self) -> ty.Sequence[TypeValue]:
        self._raise_initialized()
        return self._value

    @property
    def bitstring(self) -> Bitstring:
        self._raise_initialized()
        bits = [element.bitstring for element in self._value]
        return Bitstring.join(bits)

    @property
    def accepted_type(self) -> type:
        return list

    @property
    def element_type(self) -> Type:
        return self._element_type


class MessageValue(TypeValue):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    _type: Message

    def __init__(
        self,
        model: Message,
        refinements: ty.Sequence["RefinementValue"] = None,
        skip_verification: bool = False,
        state: "MessageValue.State" = None,
    ) -> None:
        super().__init__(model)
        self._skip_verification = skip_verification
        self._refinements = refinements or []
        self._path: ty.List[Link] = []

        self._fields: ty.Mapping[str, MessageValue.Field] = (
            state.fields
            if state and state.fields
            else {
                f.name: self.Field(
                    TypeValue.construct(
                        self._type.types[f] if f in self._type.types else Opaque(),
                        imported=f in self._type.types
                        and self._type.types[f].package != model.package,
                    ),
                    f.name,
                )
                for f in (INITIAL,) + self._type.fields
            }
        )

        self._checksums: ty.Mapping[str, MessageValue.Checksum] = (
            state.checksums
            if state and state.checksums
            else {
                str(field_name): MessageValue.Checksum(str(field_name), parameters)
                for field_name, parameters in self._type.checksums.items()
            }
        )

        self.__type_literals: ty.Mapping[Name, Expr] = (
            state.type_literals
            if state and state.type_literals
            else {
                k: v
                for t in (
                    f.typeval.literals
                    for f in self._fields.values()
                    if isinstance(f.typeval, EnumValue)
                )
                for k, v in t.items()
            }
        )
        self.__additional_enum_literals: ty.Dict[Name, Expr] = {}
        self.__message_first_name = First("Message")
        initial = self._fields[INITIAL.name]
        initial.first = Number(0)
        initial.typeval.assign(bytes())
        self._simplified_mapping: ty.Dict[Name, Expr] = dict.fromkeys(
            [initial.name_size, initial.name_last, initial.name_first, self.__message_first_name],
            Number(0),
        )
        self.accessible_fields: ty.List[str] = []
        if self._skip_verification:
            self._last_field = INITIAL.name
        else:
            self._preset_fields(INITIAL.name)
        self.__message_last_name = Last("Message")
        self.__message_size_name = Size("Message")

    def add_refinement(self, refinement: "RefinementValue") -> None:
        self._refinements = [*(self._refinements or []), refinement]

    def clone(self) -> "MessageValue":
        return MessageValue(
            self._type,
            self._refinements,
            self._skip_verification,
            MessageValue.State(
                {
                    k: MessageValue.Field(
                        v.typeval.clone(),
                        k,
                        v.name_variable,
                        v.name_first,
                        v.name_last,
                        v.name_size,
                    )
                    for k, v in self._fields.items()
                },
                self._checksums,
                self.__type_literals,
            ),
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self._fields == other._fields and self._type == other._type
        return NotImplemented

    def equal_type(self, other: Type) -> bool:
        return self.identifier == other.identifier

    @property
    def model(self) -> Message:
        return self._type

    @property
    def path(self) -> ty.Sequence[Link]:
        """
        The returned path is only correct, if the parse() method is used to set the field values.
        If fields are set manually using the set() method, path is empty. If the set() method is
        used to manipulate fields that have already been set by parse(), path is incorrect.
        """
        return self._path

    def inner_messages(self) -> ty.List["MessageValue"]:
        messages: ty.List["MessageValue"] = []
        for field in self._fields.values():
            typeval = field.typeval
            if isinstance(typeval, SequenceValue) and isinstance(typeval.element_type, Message):
                assert isinstance(typeval.value, list)
                messages.extend(typeval.value)
            if isinstance(typeval, OpaqueValue) and typeval.nested_message is not None:
                messages.append(typeval.nested_message)
        return messages

    def _valid_refinement_condition(self, refinement: "RefinementValue") -> bool:
        return self.__simplified(refinement.condition) == TRUE

    def _next_link(self, source_field_name: str) -> ty.Optional[Link]:
        field = Field(source_field_name)
        if field == FINAL or not self._type.structure:
            # structure is empty in case of NULL message
            # ISSUE: Componolit/RecordFlux#643
            return None
        for link in self._type.outgoing(field):
            if self.__simplified(link.condition) == TRUE:
                return link
        return None

    def _next_field(self, field_name: str, append_to_path: bool = False) -> str:
        if self._skip_verification and field_name != FINAL.name and self._fields[field_name].next:
            return self._fields[field_name].next

        next_link = self._next_link(field_name)
        if append_to_path and next_link is not None:
            self._path.append(next_link)

        if next_link is None and field_name == INITIAL.name:
            # the INITIAL field has no outgoing links in case of NULL message
            # ISSUE: Componolit/RecordFlux#643
            return FINAL.name
        return next_link.target.name if next_link is not None else ""

    def _prev_field(self, fld: str) -> str:
        if fld == INITIAL.name:
            return ""
        if self._skip_verification:
            return self._fields[fld].prev
        prev: ty.List[str] = [
            l.source.name
            for l in self._type.incoming(Field(fld))
            if self.__simplified(l.condition) == TRUE
        ]

        if len(prev) == 1:
            return prev[0]
        for field in prev:
            if field in self.accessible_fields:
                return field
        return ""

    def _get_size(self, fld: str) -> ty.Optional[Number]:
        typeval = self._fields[fld].typeval
        if isinstance(typeval, ScalarValue):
            return typeval.size
        assert isinstance(typeval, CompositeValue)
        for l in self._type.incoming(Field(fld)):
            if (
                self._fields[l.source.name].set
                and l.size != UNDEFINED
                and (self._skip_verification or self.__simplified(l.condition) == TRUE)
            ):
                size = self.__simplified(l.size)
                return size if isinstance(size, Number) else None
        return None

    def _get_first(self, fld: str) -> ty.Optional[Number]:
        for l in self._type.incoming(Field(fld)):
            if l.first != UNDEFINED and (
                self._skip_verification or self.__simplified(l.condition) == TRUE
            ):
                first = self.__simplified(l.first)
                return first if isinstance(first, Number) else None
        prv = self._prev_field(fld)
        if self._skip_verification and prv:
            first = self._fields[prv].first
            size = self._fields[prv].typeval.size
            assert isinstance(first, Number)
            assert isinstance(size, Number)
            return first + size
        if prv and UNDEFINED not in (self._fields[prv].first, self._fields[prv].typeval.size):
            first = self.__simplified(Add(self._fields[prv].first, self._fields[prv].typeval.size))
            return first if isinstance(first, Number) else None
        return None

    @property
    def accepted_type(self) -> type:
        return bytes

    @property
    def size(self) -> Number:
        return Number(len(self.bitstring))

    def assign(self, value: bytes, check: bool = True) -> None:
        raise NotImplementedError

    def parse(self, value: ty.Union[Bitstring, bytes], check: bool = True) -> None:
        assert not self._skip_verification
        self._path.clear()
        if isinstance(value, bytes):
            value = Bitstring.from_bytes(value)
        current_field_name = self._next_field(INITIAL.name, append_to_path=True)
        last_field_first_in_bitstr = current_field_first_in_bitstr = 0

        def get_current_pos_in_bitstr(field_name: str) -> int:
            # if the previous node is a virtual node i.e. has the same first as the current node
            # set the current pos in bitstring back to the first position of its predecessor
            this_first = self._fields[field_name].first
            prev_first = self._fields[self._prev_field(field_name)].first

            if not isinstance(prev_first, Number) or not isinstance(this_first, Number):
                return current_field_first_in_bitstr

            return (
                last_field_first_in_bitstr
                if prev_first.value == this_first.value
                else current_field_first_in_bitstr
            )

        def set_field_without_size(
            field_name: str, field: MessageValue.Field
        ) -> ty.Tuple[int, int]:
            last_pos_in_bitstr = current_pos_in_bitstring = get_current_pos_in_bitstr(field_name)
            assert isinstance(field.typeval, CompositeValue)
            first = self._get_first(field_name)
            assert first is not None
            field.first = first
            self._set_parsed_value(field_name, value[current_pos_in_bitstring:])
            return last_pos_in_bitstr, current_pos_in_bitstring

        def set_field_with_size(field_name: str, field_size: int) -> ty.Tuple[int, int]:
            assert isinstance(value, Bitstring)
            last_pos_in_bitstr = current_pos_in_bitstring = get_current_pos_in_bitstr(field_name)
            self._set_parsed_value(
                field_name,
                value[current_pos_in_bitstring : current_pos_in_bitstring + field_size],
            )
            current_pos_in_bitstring += field_size
            return last_pos_in_bitstr, current_pos_in_bitstring

        while current_field_name != FINAL.name:
            current_field = self._fields[current_field_name]
            size = self._get_size(current_field_name)
            if isinstance(current_field.typeval, CompositeValue) and size is None:
                (
                    last_field_first_in_bitstr,
                    current_field_first_in_bitstr,
                ) = set_field_without_size(current_field_name, current_field)

            else:
                assert size is not None
                current_field_size = size.value
                try:
                    (
                        last_field_first_in_bitstr,
                        current_field_first_in_bitstr,
                    ) = set_field_with_size(current_field_name, current_field_size)
                except IndexError:
                    raise PyRFLXError(
                        f"Bitstring representing the message is too short - "
                        f"stopped while parsing field: {current_field_name}"
                    ) from None
            current_field_name = self._next_field(current_field_name, append_to_path=True)

    def _set_unchecked(
        self, field_name: str, value: ty.Union[bytes, int, str, ty.Sequence[TypeValue]]
    ) -> None:
        field = self._fields[field_name]
        field.prev = self._last_field
        self._fields[self._last_field].next = field_name
        self._last_field = field_name
        f_first = self._get_first(field_name)
        f_size = self._get_size(field_name)
        assert isinstance(f_first, Number)
        field.first = f_first
        if isinstance(field.typeval, CompositeValue) and f_size is not None:
            field.typeval.set_expected_size(f_size)
        field.typeval.assign(value, not self._skip_verification)
        self.__update_simplified_mapping(field)
        self.accessible_fields.append(field_name)

    def _set_checked(
        self, field_name: str, value: ty.Union[bytes, int, str, ty.Sequence[TypeValue], Bitstring]
    ) -> None:
        def set_refinement(fld: MessageValue.Field, fld_name: str) -> None:
            if isinstance(fld.typeval, OpaqueValue):
                for ref in self._refinements:
                    if (
                        ref.pdu.name == self.name
                        and ref.field.name == fld_name
                        and self._valid_refinement_condition(ref)
                    ):
                        fld.typeval.set_refinement(ref.sdu)

        def check_outgoing_condition_satisfied() -> None:
            if all(
                self.__simplified(o.condition) == FALSE
                for o in self._type.outgoing(Field(field_name))
            ):
                self._fields[field_name].typeval.clear()
                raise PyRFLXError(
                    f"none of the field conditions "
                    f"{[str(o.condition) for o in self._type.outgoing(Field(field_name))]}"
                    f" for field {field_name} have been met by the assigned value: {value!s}"
                )

        if field_name in self.accessible_fields:
            field = self._fields[field_name]
            f_first = field.first
            f_size = field.typeval.size
            field_first = (
                f_first
                if self._skip_verification and isinstance(f_first, Number)
                else self._get_first(field_name)
            )
            field_size = (
                f_size
                if self._skip_verification and isinstance(f_size, Number)
                else self._get_size(field_name)
            )
            assert field_first is not None
            field.first = field_first
            if isinstance(field.typeval, CompositeValue) and field_size is not None:
                field.typeval.set_expected_size(field_size)
            set_refinement(field, field_name)
            try:
                if isinstance(value, Bitstring):
                    field.typeval.parse(value)
                    if (
                        isinstance(field.typeval, EnumValue)
                        and Variable(field.typeval.value) not in self.__type_literals
                    ):
                        self.__additional_enum_literals[
                            Variable(field.typeval.value)
                        ] = field.typeval.numeric_value
                elif isinstance(value, field.typeval.accepted_type):
                    field.typeval.assign(value)
                else:
                    raise PyRFLXError(
                        f"cannot assign different types: {field.typeval.accepted_type.__name__}"
                        f" != {type(value).__name__}"
                    )
            except PyRFLXError as e:
                e.appendleft(
                    f"cannot set value for field {field_name}",
                    Subsystem.PYRFLX,
                    Severity.ERROR,
                )
                raise e
        else:
            raise PyRFLXError(f"cannot access field {field_name}")

        self.__update_simplified_mapping()
        check_outgoing_condition_satisfied()

    def set(
        self,
        field_name: str,
        value: ty.Union[bytes, int, str, ty.Sequence[TypeValue]],
    ) -> None:
        if self._skip_verification:
            self._set_unchecked(field_name, value)
            return
        self._set_checked(field_name, value)
        self._preset_fields(field_name)
        for checksum in self._checksums.values():
            if (
                not self._fields[checksum.field_name].set or checksum.calculated
            ) and self._is_checksum_settable(checksum):
                self._fields[checksum.field_name].typeval.assign(0)
                checksum.calculated = True
                checksum_value = self._calculate_checksum(checksum)
                self._set_checked(checksum.field_name, checksum_value)

    def _set_parsed_value(
        self,
        field_name: str,
        value: ty.Union[bytes, int, str, ty.Sequence[TypeValue], Bitstring],
    ) -> None:
        self._set_checked(field_name, value)
        self._preset_fields(field_name)

    def _preset_fields(self, fld: str) -> None:
        assert not self._skip_verification
        nxt = self._next_field(fld)
        fields: ty.List[str] = []
        while nxt and nxt != FINAL.name:
            field = self._fields[nxt]
            first = self._get_first(nxt)
            size = self._get_size(nxt)
            if first is None:
                break

            if (self.__simplified(self._type.field_condition(Field(nxt))) == TRUE) and (
                self._is_valid_composite_field(nxt)
                if isinstance(self._fields[nxt].typeval, CompositeValue)
                else size is not None
            ):
                fields.append(nxt)

            if size is None:
                break

            field.first = first
            if isinstance(field.typeval, OpaqueValue):
                field.typeval.set_expected_size(size)

            if field.set and isinstance(field.typeval, OpaqueValue):
                field.first = UNDEFINED
                field.typeval.clear()
                break
            self._last_field = nxt
            nxt = self._next_field(nxt)
        try:
            self.accessible_fields = (
                self.accessible_fields[: self.accessible_fields.index(fld) + 1] + fields
            )
        except ValueError:
            self.accessible_fields = fields

    def set_checksum_function(self, checksums: ty.Dict[str, ty.Callable]) -> None:
        for checksum_field_name, checksum_function in checksums.items():
            if checksum_field_name not in self.fields:
                raise PyRFLXError(
                    f"cannot set checksum function: field {checksum_field_name} is not defined"
                )
            for field_name, checksum in self._checksums.items():
                if field_name == checksum_field_name:
                    checksum.function = checksum_function
                else:
                    raise PyRFLXError(
                        f"cannot set checksum function: field {checksum_field_name} "
                        f"has not been defined as a checksum field"
                    )

    def _is_checksum_settable(self, checksum: "MessageValue.Checksum") -> bool:
        def valid_path(value_range: ValueRange) -> bool:
            lower = value_range.lower.substituted(
                func=lambda e: self._fields[self._next_field(INITIAL.name)].name_first
                if e == self.__message_first_name
                else e
            )
            expr: ty.Dict[Expr, str] = dict.fromkeys([lower, value_range.upper])

            for e in expr:
                if isinstance(e, Sub):
                    assert isinstance(e.left, (First, Last))
                    expr[e] = str(e.left.prefix)
                elif isinstance(e, Add):
                    for t in e.terms:
                        if isinstance(t, (First, Last)):
                            expr[e] = str(t.prefix)
                else:
                    assert isinstance(e, (First, Last))
                    expr[e] = str(e.prefix)

            field = expr.get(lower)
            assert isinstance(field, str)
            upper_field_name = expr[value_range.upper]
            if upper_field_name == "Message":
                upper_field_name = "Final"
            while field != upper_field_name:
                field = self._next_field(field)
                if field == "Final" or field in self._checksums:
                    continue
                if field == "" or not self._fields[field].set:
                    break
            else:
                return True

            return False

        for expr_tuple in checksum.parameters:
            expr_tuple.evaluated_expression = self.__simplified(expr_tuple.expression)
            if (
                isinstance(expr_tuple.evaluated_expression, ValueRange)
                and isinstance(expr_tuple.expression, ValueRange)
                and (
                    not isinstance(expr_tuple.evaluated_expression.lower, Number)
                    or not isinstance(expr_tuple.evaluated_expression.upper, Number)
                    or not valid_path(expr_tuple.expression)
                )
            ):
                return False
            if (
                isinstance(expr_tuple.evaluated_expression, Variable)
                and not self._fields[expr_tuple.evaluated_expression.name].set
            ):
                return False
            if (
                isinstance(expr_tuple.evaluated_expression, Attribute)
                and not self._fields[str(expr_tuple.evaluated_expression.prefix)].set
            ):
                return False
        return True

    def update_checksums(self) -> None:
        for checksum in self._checksums.values():
            self._simplified_mapping[ValidChecksum(checksum.field_name)] = TRUE
            self._is_checksum_settable(checksum)
            checksum_value = self._calculate_checksum(checksum)
            self._fields[checksum.field_name].typeval.assign(checksum_value)

    def _calculate_checksum(
        self, checksum: "MessageValue.Checksum"
    ) -> ty.Union[bytes, int, str, ty.Sequence[TypeValue], Bitstring]:
        if not checksum.function:
            raise PyRFLXError(
                f"cannot calculate checksum for {checksum.field_name}: "
                f"no callable checksum function provided"
            )

        arguments: ty.Dict[str, ty.Union[int, ty.Tuple[int, int]]] = {}
        for expr_tuple in checksum.parameters:
            if isinstance(expr_tuple.evaluated_expression, ValueRange):
                assert isinstance(expr_tuple.evaluated_expression.lower, Number) and isinstance(
                    expr_tuple.evaluated_expression.upper, Number
                )
                arguments[str(expr_tuple.expression)] = (
                    expr_tuple.evaluated_expression.lower.value,
                    expr_tuple.evaluated_expression.upper.value,
                )
            elif isinstance(expr_tuple.evaluated_expression, Variable):
                assert (
                    expr_tuple.evaluated_expression.name in self.fields
                    and self._fields[expr_tuple.evaluated_expression.name].set
                )
                arguments[str(expr_tuple.expression)] = self._fields[
                    expr_tuple.evaluated_expression.name
                ].typeval.value
            else:
                assert isinstance(expr_tuple.evaluated_expression, Number)
                arguments[str(expr_tuple.expression)] = expr_tuple.evaluated_expression.value
        return checksum.function(self._unchecked_bytestring(), **arguments)

    def get(
        self, field_name: str
    ) -> ty.Union["MessageValue", ty.Sequence[TypeValue], int, str, bytes]:
        if field_name not in self.valid_fields:
            if field_name not in self.fields:
                raise PyRFLXError(f'"{field_name}" is not a field of this message')
            raise PyRFLXError(f'"{field_name}" is not set')
        field = self._fields[field_name]
        if isinstance(field.typeval, OpaqueValue) and field.typeval.nested_message is not None:
            return field.typeval.nested_message
        return self._fields[field_name].typeval.value

    @property
    def bitstring(self) -> Bitstring:
        bits = ""
        field = self._next_field(INITIAL.name)
        while field and field != FINAL.name:
            field_val = self._fields[field]
            if (
                not field_val.set
                or not isinstance(field_val.first, Number)
                or not field_val.first.value <= len(bits)
            ):
                break
            bits = f"{bits[: field_val.first.value]}{str(self._fields[field].typeval.bitstring)}"
            field = self._next_field(field)

        return Bitstring(bits)

    @property
    def value(self) -> ty.Any:
        raise NotImplementedError

    def _unchecked_bytestring(self) -> bytes:
        bits = str(self.bitstring)
        assert len(bits) % 8 == 0
        return b"".join(
            [int(bits[i : i + 8], 2).to_bytes(1, "big") for i in range(0, len(bits), 8)]
        )

    @property
    def bytestring(self) -> bytes:
        if not self._skip_verification and not self.valid_message:
            raise PyRFLXError(f"cannot create bytestring of invalid message: {self.identifier}")
        return self._unchecked_bytestring()

    @property
    def fields(self) -> ty.List[str]:
        return [f.name for f in self._type.fields]

    def _is_valid_composite_field(self, field: str) -> bool:
        assert isinstance(self._fields[field].typeval, CompositeValue)
        incoming = self._type.incoming(Field(field))

        for l in incoming:
            if (
                l.size != UNDEFINED
                and self._fields[l.source.name].set
                and self.__simplified(l.condition) == TRUE
            ):
                valid_edge = l
                break
        else:
            return False

        return all(
            (v.name in self._fields and self._fields[v.name].set) or v.name == "Message"
            for v in valid_edge.size.variables()
        )

    @property
    def valid_fields(self) -> ty.List[str]:
        return [
            f
            for f in self.accessible_fields
            if (
                self._fields[f].set
                and self.__simplified(self._type.field_condition(Field(f))) == TRUE
                and any(
                    self.__simplified(o.condition) == TRUE for o in self._type.outgoing(Field(f))
                )
            )
        ]

    @property
    def required_fields(self) -> ty.List[str]:
        accessible = self.accessible_fields
        valid = self.valid_fields
        return [f for f in accessible if f not in valid]

    @property
    def valid_message(self) -> bool:
        return (
            bool(self.valid_fields)
            and self._next_field(self.valid_fields[-1]) == FINAL.name
            and all(
                (self._is_checksum_settable(checksum) or self._skip_verification)
                and self._calculate_checksum(checksum) == self.get(checksum.field_name)
                for checksum in self._checksums.values()
            )
        )

    def __update_simplified_mapping(self, field: ty.Optional[Field] = None) -> None:
        if field:
            if isinstance(field.typeval, ScalarValue):
                self._simplified_mapping[field.name_variable] = field.typeval.expr
            last = field.last
            assert isinstance(last, Number)
            self._simplified_mapping[field.name_size] = field.typeval.size
            self._simplified_mapping[field.name_first] = field.first
            self._simplified_mapping[field.name_last] = last
            self._simplified_mapping[self.__message_last_name] = last
            self._simplified_mapping[self.__message_size_name] = last + Number(1)
            return

        self._simplified_mapping = {
            Size(field_type.full_name): field_type.size
            for field_type in self._type.types.values()
            if isinstance(field_type, Scalar)
        }
        self._simplified_mapping[self.__message_first_name] = Number(0)
        for v in self._fields.values():
            if isinstance(v.typeval, ScalarValue) and v.set:
                self._simplified_mapping[v.name_variable] = v.typeval.expr
            if isinstance(v.typeval, ScalarValue) or v.set:
                self._simplified_mapping[v.name_size] = v.typeval.size
            if isinstance(v.first, Number):
                self._simplified_mapping[v.name_first] = v.first
            if isinstance(v.last, Number):
                self._simplified_mapping[v.name_last] = v.last

        nxt = self._next_field(INITIAL.name)
        while nxt:
            last_field = nxt
            nxt = self._next_field(last_field)
            last = self._fields[last_field].last
            if nxt == FINAL.name or not isinstance(last, Number):
                break
            if nxt:
                continue
            if any(l.target == FINAL for l in self._type.outgoing(Field(last_field))):
                self._simplified_mapping[self.__message_last_name] = last
                self._simplified_mapping[self.__message_size_name] = last + Number(1)

        # ISSUE: Componolit/RecordFlux#422
        self._simplified_mapping.update({ValidChecksum(f): TRUE for f in self._checksums})

    def __simplified(self, expr: Expr) -> Expr:
        if expr in {TRUE, FALSE}:
            return expr

        def subst(expression: Expr) -> Expr:
            if expression in self._simplified_mapping:
                assert isinstance(expression, Name)
                return self._simplified_mapping[expression]
            if expression in self.__type_literals:
                assert isinstance(expression, Name)
                return self.__type_literals[expression]
            if expression in self.__additional_enum_literals:
                assert isinstance(expression, Name)
                return self.__additional_enum_literals[expression]
            return expression

        return expr.substituted(func=subst).substituted(func=subst).simplified()

    class Checksum:
        def __init__(self, field_name: str, parameters: ty.Sequence[Expr]):
            self.field_name = field_name
            self.function: ty.Optional[ty.Callable] = None
            self.calculated = False

            @dataclass
            class ExpressionTuple:
                expression: Expr
                evaluated_expression: Expr = UNDEFINED

            self.parameters: ty.List[ExpressionTuple] = []
            for expr in parameters:
                assert isinstance(expr, (ValueRange, Attribute, Variable))
                self.parameters.append(ExpressionTuple(expr))

    @dataclass
    class Field(Base):
        typeval: TypeValue
        __first: Expr
        __last: Expr
        name_variable: Variable
        name_first: First
        name_last: Last
        name_size: Size
        prev: str
        next: str

        def __init__(
            self,
            t: TypeValue,
            name: str = "",
            name_variable: Variable = None,
            name_first: First = None,
            name_last: Last = None,
            name_size: Size = None,
        ):
            # pylint: disable=too-many-arguments
            assert name or (name_variable and name_first and name_last and name_size)
            self.typeval = t
            self.__is_scalar = isinstance(self.typeval, ScalarValue)
            self.first: Expr = UNDEFINED
            self.name_variable = name_variable if name_variable else Variable(name)
            self.name_first = name_first if name_first else First(name)
            self.name_last = name_last if name_last else Last(name)
            self.name_size = name_size if name_size else Size(name)
            self.prev = ""
            self.next = ""

        def _last(self) -> Expr:
            if self.first == UNDEFINED:
                return UNDEFINED
            return Sub(Add(self.__first, self.typeval.size), Number(1)).simplified()

        @property
        def first(self) -> Expr:
            return self.__first

        @first.setter
        def first(self, first: Expr) -> None:
            self.__first = first
            if self.__is_scalar:
                self.__last = self._last()

        def __eq__(self, other: object) -> bool:
            if isinstance(other, MessageValue.Field):
                return (
                    self.first == other.first
                    and self.last == other.last
                    and self.typeval == other.typeval
                )
            return NotImplemented

        @property
        def set(self) -> bool:
            return (
                self.typeval.initialized
                and isinstance(self.typeval.size, Number)
                and isinstance(self.first, Number)
                and isinstance(self.last, Number)
            )

        @property
        def last(self) -> Expr:
            return self.__last if self.__is_scalar else self._last()

    @dataclass
    class State:
        fields: ty.Optional[ty.Mapping[str, "MessageValue.Field"]] = None
        checksums: ty.Optional[ty.Mapping[str, "MessageValue.Checksum"]] = None
        type_literals: ty.Optional[ty.Mapping[Name, Expr]] = None


class RefinementValue:
    def __init__(self, refinement: Refinement, sdu_message: MessageValue) -> None:
        self.package = refinement.package
        self.pdu = refinement.pdu
        self.field = refinement.field
        self.sdu = sdu_message
        self.condition = refinement.condition



# pylint: disable=too-many-lines

import logging
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Mapping, Optional, Sequence, Set, Tuple, Type, Union

from librflxlang import (
    AnalysisContext,
    Aspect,
    ChecksumAspect,
    Components,
    Description,
    Diagnostic,
    EnumerationTypeDef,
    Expr,
    FormalChannelDecl,
    FormalFunctionDecl,
    FormalPrivateTypeDecl,
    GrammarRule,
    MessageTypeDef,
    ModularTypeDef,
    NullID,
    PackageNode,
    RangeTypeDef,
    RefinementDecl,
    RenamingDecl,
    RFLXNode,
    SequenceTypeDef,
    SessionDecl,
    Specification,
    State,
    Statement,
    ThenNode,
    Transition,
    TypeDecl,
    TypeDerivationDef,
    VariableDecl,
)

import rflx.declaration as decl
import rflx.expression as expr
import rflx.model as model
import rflx.statement as stmt
from rflx.error import Location, RecordFluxError, Severity, Subsystem, fail
from rflx.identifier import ID, StrID
from rflx.specification.const import RESERVED_WORDS

from .cache import Cache

log = logging.getLogger(__name__)
STDIN = Path("<stdin>")


def node_location(node: RFLXNode, filename: Path) -> Location:
    start = node.token_start.sloc_range
    end = node.token_end.sloc_range
    return Location(
        start=(start.start.line, start.start.column),
        source=filename,
        end=(end.end.line, end.end.column),
    )


def type_location(identifier: ID, node: RFLXNode) -> Location:
    """
    Create a location object covering the area from the start of an identifier to the end of a node.
    """
    assert identifier.location is not None
    assert identifier.location.source is not None
    return Location(
        identifier.location.start,
        identifier.location.source,
        node_location(node, identifier.location.source).end,
    )


def diagnostics_to_error(
    diagnostics: List[Diagnostic], error: RecordFluxError, filename: Path
) -> bool:
    """
    Append langkit diagnostics to RecordFlux error. Return True if error occured.
    """

    if len(diagnostics) == 0:
        return False

    for diag in diagnostics:
        loc = diag.sloc_range
        error.append(
            diag.message,
            Subsystem.PARSER,
            Severity.ERROR,
            Location(
                start=(loc.start.line, loc.start.column),
                source=filename,
                end=(loc.end.line, loc.end.column),
            ),
        )
    return True


def create_description(description: Description = None) -> Optional[str]:
    if description:
        return description.text.split('"')[1]
    return None


def create_transition(transition: Transition, filename: Path) -> model.Transition:
    if transition.kind_name not in ("Transition", "ConditionalTransition"):
        raise NotImplementedError(f"Transition kind {transition.kind_name} unsupported")
    target = create_id(transition.f_target, filename)
    condition: expr.Expr = expr.TRUE
    description = create_description(transition.f_description)
    if transition.kind_name == "ConditionalTransition":
        condition = create_bool_expression(transition.f_condition, filename)
    return model.Transition(target, condition, description, node_location(transition, filename))


def create_reset(reset: Statement, filename: Path) -> stmt.Statement:
    return stmt.Reset(
        create_id(reset.f_identifier, filename), location=node_location(reset, filename)
    )


def create_assignment(assignment: Statement, filename: Path) -> stmt.Statement:
    return stmt.Assignment(
        create_id(assignment.f_identifier, filename),
        create_expression(assignment.f_expression, filename),
        location=node_location(assignment, filename),
    )


def create_attribute_statement(expression: Statement, filename: Path) -> stmt.Statement:
    attrs = {
        "Append": stmt.Append,
        "Extend": stmt.Extend,
        "Read": stmt.Read,
        "Write": stmt.Write,
    }
    constructor = attrs[expression.f_attr.text]

    return constructor(
        create_id(expression.f_identifier, filename),
        create_expression(expression.f_expression, filename),
        location=node_location(expression, filename),
    )


def create_statement(statement: Statement, filename: Path) -> stmt.Statement:
    handlers = {
        "Reset": create_reset,
        "Assignment": create_assignment,
        "AttributeStatement": create_attribute_statement,
    }
    return handlers[statement.kind_name](statement, filename)


def create_state(state: State, filename: Path) -> model.State:
    location = node_location(state, filename)
    identifier = create_id(state.f_identifier, filename)
    if state.f_body.kind_name == "NullStateBody":
        return model.State(identifier)
    if state.f_identifier.text != state.f_body.f_end_identifier.text:
        fail(
            "inconsistent state identifier: "
            f"{state.f_identifier.text} /= {state.f_body.f_end_identifier.text}",
            Subsystem.PARSER,
            Severity.ERROR,
            location,
        )
    transitions = []
    for t in state.f_body.f_conditional_transitions:
        transitions.append(create_transition(t, filename))
    transitions.append(create_transition(state.f_body.f_final_transition, filename))
    exception_transition = (
        create_transition(state.f_body.f_exception_transition, filename)
        if state.f_body.f_exception_transition
        else None
    )
    actions = []
    for a in state.f_body.f_actions:
        actions.append(create_statement(a, filename))
    declarations = []
    for d in state.f_body.f_declarations:
        declarations.append(create_declaration(d, filename))
    description = create_description(state.f_description)
    return model.State(
        identifier=identifier,
        transitions=transitions,
        exception_transition=exception_transition,
        actions=actions,
        declarations=declarations,
        description=description,
        location=node_location(state, filename),
    )


def __check_session_identifier(session: SessionDecl, filename: Path) -> None:
    if session.f_identifier.text != session.f_end_identifier.text:
        fail(
            "inconsistent session identifier: "
            f"{session.f_identifier.text} /= {session.f_end_identifier.text}",
            Subsystem.PARSER,
            Severity.ERROR,
            node_location(session, filename),
        )


def create_unproven_session(
    session: SessionDecl,
    package: ID,
    filename: Path,
    types: Sequence[model.Type] = None,
) -> model.UnprovenSession:
    __check_session_identifier(session, filename)
    return model.UnprovenSession(
        package * create_id(session.f_identifier, filename),
        create_id(session.f_aspects.f_initial, filename),
        create_id(session.f_aspects.f_final, filename),
        [create_state(s, filename) for s in session.f_states],
        [create_declaration(d, filename) for d in session.f_declarations],
        [create_formal_declaration(p, filename) for p in session.f_parameters],
        types or [],
        node_location(session, filename),
    )


def create_session(
    session: SessionDecl,
    package: ID,
    filename: Path,
    types: Sequence[model.Type] = None,
) -> model.Session:
    return create_unproven_session(session, package, filename, types).proven()


def create_id(identifier: NullID, filename: Path) -> ID:
    if identifier.kind_name == "UnqualifiedID":
        if identifier.text.lower() in RESERVED_WORDS:
            fail(
                f'reserved word "{identifier.text}" used as identifier',
                Subsystem.PARSER,
                Severity.ERROR,
                node_location(identifier, filename),
            )
        return ID(identifier.text, location=node_location(identifier, filename))
    if identifier.kind_name == "ID":
        name = ID(identifier.f_name.text, location=node_location(identifier.f_name, filename))
        if identifier.f_package:
            return (
                ID(
                    identifier.f_package.text,
                    location=node_location(identifier.f_package, filename),
                )
                * name
            )
        return name

    raise NotImplementedError(f"Invalid ID: {identifier.text}")


def create_sequence(
    identifier: ID,
    sequence: SequenceTypeDef,
    types: Sequence[model.Type],
    _skip_verification: bool,
    _cache: Cache,
    filename: Path,
) -> model.Sequence:
    element_identifier = model.qualified_type_identifier(
        create_id(sequence.f_element_type, filename), identifier.parent
    )

    try:
        element_type = next(t for t in types if element_identifier == t.identifier)
    except StopIteration:
        fail(
            f'undefined element type "{element_identifier}"',
            Subsystem.PARSER,
            Severity.ERROR,
            element_identifier.location,
        )

    return model.Sequence(identifier, element_type, type_location(identifier, sequence))


def create_numeric_literal(expression: Expr, filename: Path) -> expr.Expr:
    location = node_location(expression, filename)
    num = expression.text.split("#")
    if len(num) == 1:
        return expr.Number(int(num[0]), location=location)
    if len(num) == 3:
        base = int(num[0])
        return expr.Number(int(num[1], base), base=base, location=location)
    raise NotImplementedError(f"Invalid numeric literal: {expression.text}")


OPERATIONS: Dict[str, Type[expr.BinExpr]] = {
    "OpIn": expr.In,
    "OpNotin": expr.NotIn,
    "OpEq": expr.Equal,
    "OpNeq": expr.NotEqual,
}


def create_binop(expression: Expr, filename: Path) -> expr.Expr:
    loc = node_location(expression, filename)
    if expression.f_op.kind_name in OPERATIONS:
        return OPERATIONS[expression.f_op.kind_name](
            create_expression(expression.f_left, filename),
            create_expression(expression.f_right, filename),
            location=loc,
        )
    if expression.f_op.kind_name in BOOLEAN_OPERATIONS:
        return BOOLEAN_OPERATIONS[expression.f_op.kind_name](
            create_expression(expression.f_left, filename),
            create_expression(expression.f_right, filename),
            location=loc,
        )

    left = create_math_expression(expression.f_left, filename)
    right = create_math_expression(expression.f_right, filename)
    if expression.f_op.kind_name in MATH_OPERATIONS:
        return MATH_OPERATIONS[expression.f_op.kind_name](left, right, location=loc)
    if expression.f_op.kind_name in MATH_COMPARISONS:
        return MATH_COMPARISONS[expression.f_op.kind_name](left, right, location=loc)

    raise NotImplementedError(f"Invalid BinOp {expression.f_op.kind_name} => {expression.text}")


MATH_OPERATIONS = {
    "OpPow": expr.Pow,
    "OpAdd": expr.Add,
    "OpSub": expr.Sub,
    "OpMul": expr.Mul,
    "OpDiv": expr.Div,
    "OpMod": expr.Mod,
}


def create_math_binop(expression: Expr, filename: Path) -> expr.Expr:
    if expression.f_op.kind_name in MATH_OPERATIONS:
        return MATH_OPERATIONS[expression.f_op.kind_name](
            create_math_expression(expression.f_left, filename),
            create_math_expression(expression.f_right, filename),
            location=node_location(expression, filename),
        )
    raise NotImplementedError(
        f"Invalid math BinOp {expression.f_op.kind_name} => {expression.text}"
    )


MATH_COMPARISONS: Dict[str, Type[expr.Relation]] = {
    "OpLt": expr.Less,
    "OpGt": expr.Greater,
    "OpLe": expr.LessEqual,
    "OpGe": expr.GreaterEqual,
}

BOOLEAN_OPERATIONS = {
    "OpAnd": expr.And,
    "OpOr": expr.Or,
}


def create_bool_binop(expression: Expr, filename: Path) -> expr.Expr:
    if expression.f_op.kind_name in MATH_COMPARISONS:
        return MATH_COMPARISONS[expression.f_op.kind_name](
            create_math_expression(expression.f_left, filename),
            create_math_expression(expression.f_right, filename),
            location=node_location(expression, filename),
        )
    if expression.f_op.kind_name in BOOLEAN_OPERATIONS:
        return BOOLEAN_OPERATIONS[expression.f_op.kind_name](
            create_bool_expression(expression.f_left, filename),
            create_bool_expression(expression.f_right, filename),
            location=node_location(expression, filename),
        )
    if expression.f_op.kind_name in OPERATIONS:
        return OPERATIONS[expression.f_op.kind_name](
            create_expression(expression.f_left, filename),
            create_expression(expression.f_right, filename),
            location=node_location(expression, filename),
        )
    raise NotImplementedError(
        f"Invalid bool BinOp {expression.f_op.kind_name} => {expression.text}"
    )


def create_paren_bool_expression(expression: Expr, filename: Path) -> expr.Expr:
    return create_bool_expression(expression.f_data, filename)


def create_paren_math_expression(expression: Expr, filename: Path) -> expr.Expr:
    return create_math_expression(expression.f_data, filename)


def create_paren_expression(expression: Expr, filename: Path) -> expr.Expr:
    return create_expression(expression.f_data, filename)


def create_variable(expression: Expr, filename: Path) -> expr.Expr:
    location = node_location(expression, filename)
    if expression.f_identifier.text.lower() in ("true", "false"):
        return expr.Variable(create_id(expression.f_identifier, filename), location=location)
    return expr.Variable(create_id(expression.f_identifier, filename), location=location)


def create_math_attribute(expression: Expr, filename: Path) -> expr.Expr:
    inner = create_expression(expression.f_expression, filename)
    if expression.f_kind.kind_name == "AttrLast":
        return expr.Last(inner)
    if expression.f_kind.kind_name == "AttrFirst":
        return expr.First(inner)
    if expression.f_kind.kind_name == "AttrSize":
        return expr.Size(inner)
    raise NotImplementedError(
        f"Invalid math attribute: {expression.f_kind.kind_name} => {expression.text}"
    )


def create_attribute(expression: Expr, filename: Path) -> expr.Expr:
    inner = create_expression(expression.f_expression, filename)
    if expression.f_kind.kind_name == "AttrLast":
        return expr.Last(inner)
    if expression.f_kind.kind_name == "AttrFirst":
        return expr.First(inner)
    if expression.f_kind.kind_name == "AttrSize":
        return expr.Size(inner)
    if expression.f_kind.kind_name == "AttrValidChecksum":
        return expr.ValidChecksum(inner)
    if expression.f_kind.kind_name == "AttrHasData":
        return expr.HasData(inner)
    if expression.f_kind.kind_name == "AttrHead":
        return expr.Head(inner)
    if expression.f_kind.kind_name == "AttrOpaque":
        return expr.Opaque(inner)
    if expression.f_kind.kind_name == "AttrPresent":
        return expr.Present(inner)
    if expression.f_kind.kind_name == "AttrValid":
        return expr.Valid(inner)
    raise NotImplementedError(
        f"Invalid attribute: {expression.f_kind.kind_name} => {expression.text}"
    )


def create_sequence_aggregate(expression: Expr, filename: Path) -> expr.Expr:
    return expr.Aggregate(
        *[create_math_expression(v, filename) for v in expression.f_values],
        location=node_location(expression, filename),
    )


def create_string_literal(expression: Expr, filename: Path) -> expr.Expr:
    return expr.String(
        expression.text.split('"')[1],
        location=node_location(expression, filename),
    )


def create_call(expression: Expr, filename: Path) -> expr.Expr:
    return expr.Call(
        create_id(expression.f_identifier, filename),
        [create_expression(a, filename) for a in expression.f_arguments],
        location=node_location(expression, filename),
    )


def create_quantified_expression(expression: Expr, filename: Path) -> expr.Expr:
    param_id = create_id(expression.f_parameter_identifier, filename)
    iterable = create_expression(expression.f_iterable, filename)
    predicate = create_expression(expression.f_predicate, filename)
    location = node_location(expression, filename)
    if expression.f_operation.kind_name == "QuantifierAll":
        return expr.ForAllIn(param_id, iterable, predicate, location)
    if expression.f_operation.kind_name == "QuantifierSome":
        return expr.ForSomeIn(param_id, iterable, predicate, location)

    raise NotImplementedError(f"Invalid quantified: {expression.f_operation.text}")


def create_binding(expression: Expr, filename: Path) -> expr.Expr:
    bindings: Mapping[Union[str, ID], expr.Expr] = {
        create_id(b.f_identifier, filename): create_expression(b.f_expression, filename)
        for b in expression.f_bindings
    }
    return expr.Binding(
        create_expression(expression.f_expression, filename),
        bindings,
        node_location(expression, filename),
    )


def create_variable_decl(declaration: VariableDecl, filename: Path) -> decl.BasicDeclaration:
    initializer = (
        create_expression(declaration.f_initializer, filename)
        if declaration.f_initializer
        else None
    )
    return decl.VariableDeclaration(
        create_id(declaration.f_identifier, filename),
        model.qualified_type_identifier(create_id(declaration.f_type_identifier, filename)),
        initializer,
        location=node_location(declaration, filename),
    )


def create_private_type_decl(
    declaration: FormalPrivateTypeDecl, filename: Path
) -> decl.FormalDeclaration:
    return decl.TypeDeclaration(
        model.Private(
            create_id(declaration.f_identifier, filename),
            location=node_location(declaration, filename),
        )
    )


def create_channel_decl(
    declaration: FormalChannelDecl,
    filename: Path,
) -> decl.FormalDeclaration:
    readable = False
    writable = False
    for p in declaration.f_parameters:
        if p.kind_name == "Readable":
            readable = True
        elif p.kind_name == "Writable":
            writable = True
        else:
            raise NotImplementedError(f"channel parameter: {p.kind_name}")
    return decl.ChannelDeclaration(
        create_id(declaration.f_identifier, filename),
        readable=readable,
        writable=writable,
        location=node_location(declaration, filename),
    )


def create_renaming_decl(declaration: RenamingDecl, filename: Path) -> decl.BasicDeclaration:
    selected = create_expression(declaration.f_expression, filename)
    assert isinstance(selected, expr.Selected)
    return decl.RenamingDeclaration(
        create_id(declaration.f_identifier, filename),
        model.qualified_type_identifier(create_id(declaration.f_type_identifier, filename), None),
        selected,
        location=node_location(declaration, filename),
    )


def create_function_decl(
    declaration: FormalFunctionDecl,
    filename: Path,
) -> decl.FormalDeclaration:
    arguments = []
    if declaration.f_parameters:
        for p in declaration.f_parameters.f_parameters:
            arguments.append(
                decl.Argument(
                    create_id(p.f_identifier, filename),
                    create_id(p.f_type_identifier, filename),
                )
            )
    return decl.FunctionDeclaration(
        create_id(declaration.f_identifier, filename),
        arguments,
        create_id(declaration.f_return_type_identifier, filename),
        location=node_location(declaration, filename),
    )


def create_negation(expression: Expr, filename: Path) -> expr.Expr:
    math_expr = create_math_expression(expression.f_data, filename)
    assert isinstance(math_expr, expr.Number)
    return expr.Number(-math_expr.value, math_expr.base, node_location(expression, filename))


def create_concatenation(expression: Expr, filename: Path) -> expr.Expr:
    left = create_expression(expression.f_left, filename)
    right = create_expression(expression.f_right, filename)
    assert isinstance(left, expr.Aggregate)
    assert isinstance(right, expr.Aggregate)
    return expr.Aggregate(
        *(left.elements + right.elements), location=node_location(expression, filename)
    )


def create_comprehension(expression: Expr, filename: Path) -> expr.Expr:
    condition = (
        create_bool_expression(expression.f_condition, filename)
        if expression.f_condition
        else expr.TRUE
    )
    return expr.Comprehension(
        create_id(expression.f_iterator, filename),
        create_expression(expression.f_sequence, filename),
        create_expression(expression.f_selector, filename),
        condition,
        node_location(expression, filename),
    )


def create_selected(expression: Expr, filename: Path) -> expr.Expr:
    return expr.Selected(
        create_expression(expression.f_expression, filename),
        create_id(expression.f_selector, filename),
        location=node_location(expression, filename),
    )


def create_conversion(expression: Expr, filename: Path) -> expr.Expr:
    return expr.Conversion(
        create_id(expression.f_target_identifier, filename),
        create_expression(expression.f_argument, filename),
        location=node_location(expression, filename),
    )


def create_message_aggregate(expression: Expr, filename: Path) -> expr.Expr:
    values: Mapping[StrID, expr.Expr] = {}
    if expression.f_values.kind_name == "NullMessageAggregate":
        values = {}
    elif expression.f_values.kind_name == "MessageAggregateAssociations":
        values = {
            create_id(c.f_identifier, filename): create_expression(c.f_expression, filename)
            for c in expression.f_values.f_associations
        }
    else:
        raise NotImplementedError(f"invalid message component: {expression.f_values.kind_name}")

    return expr.MessageAggregate(
        create_id(expression.f_identifier, filename),
        values,
        location=node_location(expression, filename),
    )


EXPRESSION_MAP = {
    "NumericLiteral": create_numeric_literal,
    "ParenExpression": create_paren_expression,
    "Variable": create_variable,
    "Attribute": create_attribute,
    "SequenceAggregate": create_sequence_aggregate,
    "StringLiteral": create_string_literal,
    "Call": create_call,
    "QuantifiedExpression": create_quantified_expression,
    "Binding": create_binding,
    "Negation": create_negation,
    "Concatenation": create_concatenation,
    "Comprehension": create_comprehension,
    "SelectNode": create_selected,
    "Conversion": create_conversion,
    "MessageAggregate": create_message_aggregate,
    "BinOp": create_binop,
}


def create_expression(expression: Expr, filename: Path) -> expr.Expr:
    return EXPRESSION_MAP[expression.kind_name](expression, filename)


def create_declaration(declaration: Expr, filename: Path) -> decl.BasicDeclaration:
    handlers = {
        "VariableDecl": create_variable_decl,
        "RenamingDecl": create_renaming_decl,
    }
    return handlers[declaration.kind_name](declaration, filename)


def create_formal_declaration(declaration: Expr, filename: Path) -> decl.FormalDeclaration:
    handlers = {
        "FormalChannelDecl": create_channel_decl,
        "FormalFunctionDecl": create_function_decl,
        "FormalPrivateTypeDecl": create_private_type_decl,
    }
    return handlers[declaration.kind_name](declaration, filename)


def create_math_expression(expression: Expr, filename: Path) -> expr.Expr:
    handlers = {
        "NumericLiteral": create_numeric_literal,
        "BinOp": create_math_binop,
        "ParenExpression": create_paren_math_expression,
        "Variable": create_variable,
        "Call": create_call,
        "Binding": create_binding,
        "Negation": create_negation,
        "Attribute": create_math_attribute,
        "SelectNode": create_selected,
        "SequenceAggregate": create_sequence_aggregate,
    }
    return handlers[expression.kind_name](expression, filename)


def create_bool_expression(expression: Expr, filename: Path) -> expr.Expr:
    handlers = {
        "BinOp": create_bool_binop,
        "ParenExpression": create_paren_bool_expression,
        "Attribute": create_attribute,
        "Call": create_call,
        "Variable": create_variable,
        "QuantifiedExpression": create_quantified_expression,
        "Binding": create_binding,
        "SelectNode": create_selected,
    }
    return handlers[expression.kind_name](expression, filename)


def create_modular(
    identifier: ID,
    modular: ModularTypeDef,
    _types: Sequence[model.Type],
    _skip_verification: bool,
    _cache: Cache,
    filename: Path,
) -> model.ModularInteger:
    return model.ModularInteger(
        identifier,
        create_math_expression(modular.f_mod, filename),
        type_location(identifier, modular),
    )


def create_range(
    identifier: ID,
    rangetype: RangeTypeDef,
    _types: Sequence[model.Type],
    _skip_verification: bool,
    _cache: Cache,
    filename: Path,
) -> model.RangeInteger:
    if rangetype.f_size.f_identifier.text != "Size":
        fail(
            f"invalid aspect {rangetype.f_size.f_identifier.text} for range type {identifier}",
            Subsystem.PARSER,
            Severity.ERROR,
            node_location(rangetype, filename),
        )
    size = create_math_expression(rangetype.f_size.f_value, filename)
    return model.RangeInteger(
        identifier,
        create_math_expression(rangetype.f_first, filename),
        create_math_expression(rangetype.f_last, filename),
        size,
        type_location(identifier, rangetype),
    )


def create_null_message(
    identifier: ID,
    message: MessageTypeDef,
    _types: Sequence[model.Type],
    _skip_verification: bool,
    _cache: Cache,
    _filename: Path,
) -> model.Message:
    return model.Message(identifier, [], {}, location=type_location(identifier, message))


def create_message(
    identifier: ID,
    message: MessageTypeDef,
    types: Sequence[model.Type],
    skip_verification: bool,
    cache: Cache,
    filename: Path,
) -> model.Message:

    error = RecordFluxError()
    components = message.f_components

    field_types: Mapping[model.Field, model.Type] = create_message_types(
        identifier, components, types, filename
    )
    structure = create_message_structure(error, components, filename)
    aspects = {ID("Checksum"): create_message_aspects(message.f_checksums, filename)}

    try:
        result = create_proven_message(
            model.UnprovenMessage(
                identifier, structure, field_types, aspects, type_location(identifier, message)
            ).merged(),
            skip_verification,
            cache,
        )
    except RecordFluxError as e:
        error.extend(e)
    error.propagate()
    return result


def create_message_types(
    identifier: ID,
    components: Components,
    types: Sequence[model.Type],
    filename: Path,
) -> Mapping[model.Field, model.Type]:

    field_types: Dict[model.Field, model.Type] = {}

    for component in components.f_components:
        type_identifier = model.qualified_type_identifier(
            create_id(component.f_type_identifier, filename), identifier.parent
        )
        field_type = [t for t in types if t.identifier == type_identifier]
        if field_type:
            field_types[model.Field(create_id(component.f_identifier, filename))] = field_type[0]

    return field_types


def create_message_structure(
    error: RecordFluxError, components: Components, filename: Path
) -> List[model.Link]:
    def extract_aspect(aspects: List[Aspect]) -> Tuple[expr.Expr, expr.Expr]:
        size: expr.Expr = expr.UNDEFINED
        first: expr.Expr = expr.UNDEFINED
        for aspect in aspects:
            if aspect.f_identifier.text == "Size":
                size = create_math_expression(aspect.f_value, filename)
            elif aspect.f_identifier.text == "First":
                first = create_math_expression(aspect.f_value, filename)
            else:
                error.append(
                    f'invalid aspect "{aspect.f_identifier.text}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(aspect.f_identifier, filename),
                )
        return size, first

    def extract_then(
        then: ThenNode,
    ) -> Tuple[model.Field, expr.Expr, expr.Expr, expr.Expr, Location]:
        target = (
            model.FINAL
            if then.f_target.text == "null"
            else model.Field(create_id(then.f_target, filename))
        )
        condition = (
            create_bool_expression(then.f_condition, filename) if then.f_condition else expr.TRUE
        )
        size, first = extract_aspect(then.f_aspects)
        return target, condition, size, first, node_location(then, filename)

    structure: List[model.Link] = []

    if components.f_initial_component:
        structure.append(
            model.Link(model.INITIAL, *extract_then(components.f_initial_component.f_then))
        )
    else:
        structure.append(
            model.Link(
                model.INITIAL,
                model.Field(create_id(components.f_components[0].f_identifier, filename)),
            )
        )

    for i, component in enumerate(components.f_components):
        source_node = (
            model.Field(create_id(component.f_identifier, filename))
            if component.f_identifier
            else model.INITIAL
        )
        component_identifier = create_id(component.f_identifier, filename)
        if component.f_identifier.text.lower() == "message":
            error.append(
                'reserved word "Message" used as identifier',
                Subsystem.PARSER,
                Severity.ERROR,
                component_identifier.location,
            )
            continue

        if len(component.f_thens) == 0:
            target_id = (
                create_id(components.f_components[i + 1].f_identifier, filename)
                if i + 1 < len(components.f_components)
                else None
            )
            target_node = model.Field(target_id) if target_id else model.FINAL
            structure.append(model.Link(source_node, target_node))

        for then in component.f_thens:
            if then.f_target.kind_name != "NullID" and not any(
                then.f_target.text == c.f_identifier.text for c in components.f_components
            ):
                error.append(
                    f'undefined field "{then.f_target.text}"',
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(then.f_target, filename) if then.f_target else None,
                )
                continue
            structure.append(model.Link(source_node, *extract_then(then)))

        merge_component_aspects(
            error, component_identifier, structure, *extract_aspect(component.f_aspects)
        )
        merge_component_condition(
            component_identifier,
            structure,
            create_bool_expression(component.f_condition, filename)
            if component.f_condition
            else expr.TRUE,
        )

    return structure


def merge_component_aspects(
    error: RecordFluxError,
    component_identifier: ID,
    structure: Sequence[model.Link],
    size: Expr,
    first: Expr,
) -> None:
    if first != expr.UNDEFINED or size != expr.UNDEFINED:
        for l in (l for l in structure if l.target.identifier == component_identifier):
            if first != expr.UNDEFINED:
                if l.first == expr.UNDEFINED:
                    l.first = first
                else:
                    error.append(
                        f'first aspect of field "{component_identifier}"'
                        " conflicts with previous"
                        " specification",
                        Subsystem.MODEL,
                        Severity.ERROR,
                        first.location,
                    )
                    error.append(
                        "previous specification of first",
                        Subsystem.MODEL,
                        Severity.INFO,
                        l.first.location,
                    )

            if size != expr.UNDEFINED:
                if l.size == expr.UNDEFINED:
                    l.size = size
                else:
                    error.append(
                        f'size aspect of field "{component_identifier}" conflicts with previous'
                        " specification",
                        Subsystem.MODEL,
                        Severity.ERROR,
                        size.location,
                    )
                    error.append(
                        "previous specification of size",
                        Subsystem.MODEL,
                        Severity.INFO,
                        l.size.location,
                    )


def merge_component_condition(
    component_identifier: ID,
    structure: Sequence[model.Link],
    condition: Expr,
) -> None:
    if condition != expr.TRUE:
        for l in (l for l in structure if l.source.identifier == component_identifier):
            l.condition = (
                expr.And(condition, l.condition, location=l.condition.location)
                if l.condition != expr.TRUE
                else condition
            )


def create_message_aspects(
    checksum: ChecksumAspect, filename: Path
) -> Mapping[ID, Sequence[expr.Expr]]:
    result = {}
    if checksum:
        for assoc in checksum.f_associations:
            exprs = []
            for value in assoc.f_covered_fields:
                if value.kind_name == "ChecksumVal":
                    exprs.append(create_math_expression(value.f_data, filename))
                elif value.kind_name == "ChecksumValueRange":
                    exprs.append(
                        expr.ValueRange(
                            create_math_expression(value.f_first, filename),
                            create_math_expression(value.f_last, filename),
                        )
                    )
                else:
                    raise NotImplementedError(f"Invalid checksum association {value.kind_name}")
            result[create_id(assoc.f_identifier, filename)] = exprs
    return result


def create_derived_message(
    identifier: ID,
    derivation: TypeDerivationDef,
    types: Sequence[model.Type],
    skip_verification: bool,
    cache: Cache,
    filename: Path,
) -> model.Message:
    base_id = create_id(derivation.f_base, filename)
    base_name = model.qualified_type_identifier(base_id, identifier.parent)

    base_types: Sequence[model.Type] = [t for t in types if t.identifier == base_name]

    if not base_types:
        fail(
            f'undefined base message "{base_name}" in derived message',
            Subsystem.PARSER,
            Severity.ERROR,
            base_name.location,
        )

    base_messages: Sequence[model.Message] = [t for t in base_types if isinstance(t, model.Message)]

    if not base_messages:
        error = RecordFluxError()
        error.append(
            f'illegal derivation "{identifier}"',
            Subsystem.PARSER,
            Severity.ERROR,
            identifier.location,
        )
        error.append(
            f'invalid base message type "{base_name}"',
            Subsystem.PARSER,
            Severity.INFO,
            base_types[0].identifier.location,
        )
        error.propagate()

    return create_proven_message(
        model.UnprovenDerivedMessage(
            identifier, base_messages[0], location=type_location(identifier, derivation)
        ).merged(),
        skip_verification,
        cache,
    )


def create_enumeration(
    identifier: ID,
    enumeration: EnumerationTypeDef,
    _types: Sequence[model.Type],
    _skip_verification: bool,
    _cache: Cache,
    filename: Path,
) -> model.Enumeration:
    literals: List[Tuple[StrID, expr.Number]] = []
    error = RecordFluxError()

    def create_aspects(aspects: List[Aspect]) -> Tuple[expr.Expr, bool]:
        always_valid = False
        size = None
        for a in aspects:
            if a.f_identifier.text == "Size":
                size = create_math_expression(a.f_value, filename)
            if a.f_identifier.text == "Always_Valid":
                if a.f_value:
                    av_expr = create_bool_expression(a.f_value, filename)
                    if av_expr == expr.Variable("True"):
                        always_valid = True
                    elif av_expr == expr.Variable("False"):
                        always_valid = False
                    else:
                        error.append(
                            f"invalid Always_Valid expression: {av_expr}",
                            Subsystem.PARSER,
                            Severity.ERROR,
                            node_location(a.f_value, filename),
                        )
                else:
                    always_valid = True
        if not size:
            error.append(
                f'no size set for "{identifier}"',
                Subsystem.PARSER,
                Severity.ERROR,
                identifier.location,
            )
        error.propagate()
        assert size
        return size, always_valid

    if enumeration.f_elements.kind_name == "NamedEnumerationDef":
        for e in enumeration.f_elements.f_elements:
            element_identifier = create_id(e.f_identifier, filename)
            value = create_math_expression(e.f_literal, filename)
            assert isinstance(value, expr.Number)
            literals.append((element_identifier, value))
    elif enumeration.f_elements.kind_name == "PositionalEnumerationDef":
        literals = [
            (create_id(e, filename), expr.Number(i))
            for i, e in enumerate(enumeration.f_elements.f_elements)
        ]
    else:
        raise NotImplementedError(
            f"Enumeration kind {enumeration.f_elements.kind_name} unsupported"
        )

    size, always_valid = create_aspects(enumeration.f_aspects)

    return model.Enumeration(
        identifier, literals, size, always_valid, location=type_location(identifier, enumeration)
    )


def create_proven_message(
    unproven_message: model.UnprovenMessage, skip_verification: bool, cache: Cache
) -> model.Message:
    proven_message = unproven_message.proven(
        skip_verification or cache.is_verified(unproven_message)
    )

    cache.add_verified(unproven_message)

    return proven_message


def create_refinement(
    refinement: RefinementDecl, package: ID, types: Sequence[model.Type], filename: Path
) -> model.Refinement:
    messages = {t.identifier: t for t in types if isinstance(t, model.Message)}

    pdu = model.qualified_type_identifier(create_id(refinement.f_pdu, filename), package)
    if pdu not in messages:
        fail(
            f'undefined type "{pdu}" in refinement',
            Subsystem.PARSER,
            Severity.ERROR,
            node_location(refinement, filename),
        )

    sdu = model.qualified_type_identifier(create_id(refinement.f_sdu, filename), package)
    if sdu not in messages:
        fail(
            f'undefined type "{sdu}" in refinement of "{pdu}"',
            Subsystem.PARSER,
            Severity.ERROR,
            sdu.location,
        )

    if refinement.f_condition:
        condition = create_bool_expression(refinement.f_condition, filename)
    else:
        condition = expr.TRUE

    return model.Refinement(
        package,
        messages[pdu],
        model.Field(create_id(refinement.f_field, filename)),
        messages[sdu],
        condition,
        node_location(refinement, filename),
    )


def check_naming(error: RecordFluxError, package: PackageNode, name: Path) -> None:
    identifier = package.f_identifier.text
    if identifier.startswith("RFLX"):
        error.append(
            f'illegal prefix "RFLX" in package identifier "{identifier}"',
            Subsystem.PARSER,
            Severity.ERROR,
            node_location(package.f_identifier, name),
        )
    if identifier != package.f_end_identifier.text:
        error.append(
            f'inconsistent package identifier "{package.f_end_identifier.text}"',
            Subsystem.PARSER,
            Severity.ERROR,
            node_location(package.f_end_identifier, name),
        )
        error.append(
            f'previous identifier was "{identifier}"',
            Subsystem.PARSER,
            Severity.INFO,
            node_location(package.f_identifier, name),
        )
    if name != STDIN:
        expected_filename = f"{identifier.lower()}.rflx"
        if name.name != expected_filename:

            error.append(
                f'file name does not match unit name "{identifier}",'
                f' should be "{expected_filename}"',
                Subsystem.PARSER,
                Severity.ERROR,
                node_location(package.f_identifier, name),
            )
    for t in package.f_declarations:
        if isinstance(t, TypeDecl) and model.is_builtin_type(create_id(t.f_identifier, name).name):
            error.append(
                f'illegal redefinition of built-in type "{t.f_identifier.text}"',
                Subsystem.MODEL,
                Severity.ERROR,
                node_location(t, name),
            )


@dataclass(frozen=True)
class SpecificationNode:
    filename: Path
    spec: Specification
    withed_files: List[str]


class Parser:
    def __init__(self, skip_verification: bool = False, cached: bool = False) -> None:
        self.skip_verification = skip_verification
        self.__specifications: OrderedDict[str, SpecificationNode] = OrderedDict()
        self.__types: List[model.Type] = [
            *model.BUILTIN_TYPES.values(),
            *model.INTERNAL_TYPES.values(),
        ]
        self.__sessions: List[model.Session] = []
        self.__cache = Cache(not skip_verification and cached)

    def __convert_unit(
        self,
        error: RecordFluxError,
        spec: Specification,
        filename: Path,
        transitions: List[ID] = None,
    ) -> None:
        transitions = transitions or []
        withed_files = []

        if spec:
            check_naming(error, spec.f_package_declaration, filename)
            packagefile = f"{spec.f_package_declaration.f_identifier.text.lower()}.rflx"
            for context in spec.f_context_clause:
                item = create_id(context.f_item, filename)
                if item in transitions:
                    error.append(
                        f'dependency cycle when including "{transitions[0]}"',
                        Subsystem.PARSER,
                        Severity.ERROR,
                        transitions[0].location,
                    )
                    error.extend(
                        [
                            (
                                f'when including "{i}"',
                                Subsystem.PARSER,
                                Severity.INFO,
                                i.location,
                            )
                            for i in transitions[1:] + [item]
                        ]
                    )
                    continue
                withed_file = filename.parent / f"{str(item).lower()}.rflx"
                withed_files.append(withed_file.name)
                if withed_file.name not in self.__specifications:
                    error.extend(self.__parse_specfile(withed_file, transitions + [item]))

            if (
                packagefile in self.__specifications
                and filename != self.__specifications[packagefile].filename
            ):
                error.append(
                    "duplicate specification",
                    Subsystem.PARSER,
                    Severity.ERROR,
                    node_location(spec.f_package_declaration.f_identifier, filename),
                )
                error.append(
                    "previous specification",
                    Subsystem.PARSER,
                    Severity.INFO,
                    node_location(
                        self.__specifications[packagefile].spec.f_package_declaration.f_identifier,
                        self.__specifications[packagefile].filename,
                    ),
                )
            self.__specifications[packagefile] = SpecificationNode(filename, spec, withed_files)

    def __parse_specfile(self, filename: Path, transitions: List[ID] = None) -> RecordFluxError:
        error = RecordFluxError()
        transitions = transitions or []

        log.info("Parsing %s", filename)
        unit = AnalysisContext().get_from_file(str(filename))
        if diagnostics_to_error(unit.diagnostics, error, filename):
            return error
        self.__convert_unit(error, unit.root, filename, transitions)
        return error

    def __sort_specs_topologically(self) -> None:
        """
        (Reverse) Topologically sort specifications using Kahn's algorithm.
        """

        result: List[str] = []
        incoming: Dict[str, Set[str]] = {f: set() for f in self.__specifications.keys()}
        for filename, spec_node in self.__specifications.items():
            for d in spec_node.withed_files:
                if d in incoming:
                    incoming[d].add(filename)

        specs = [f for f, i in incoming.items() if len(i) == 0]
        visited = set(specs)

        while specs:
            s = specs.pop(0)
            result.insert(0, s)
            for e in self.__specifications[s].withed_files:
                visited.add(e)
                if e in incoming and incoming[e] <= visited:
                    specs.append(e)

        self.__specifications = OrderedDict((f, self.__specifications[f]) for f in result)

    def parse(self, *specfiles: Path) -> None:
        error = RecordFluxError()

        for f in specfiles:
            error.extend(self.__parse_specfile(f))
        self.__sort_specs_topologically()
        error.propagate()

    def parse_string(
        self,
        string: str,
        rule: GrammarRule = GrammarRule.main_rule_rule,
    ) -> None:
        error = RecordFluxError()
        unit = AnalysisContext().get_from_buffer("<stdin>", string, rule=rule)
        if not diagnostics_to_error(unit.diagnostics, error, STDIN):
            self.__convert_unit(error, unit.root, STDIN)
            self.__sort_specs_topologically()
        error.propagate()

    def create_model(self) -> model.Model:
        error = RecordFluxError()
        for spec_node in self.__specifications.values():
            self.__evaluate_specification(error, spec_node.spec, spec_node.filename)
        try:
            result = model.Model(self.__types, self.__sessions)
        except RecordFluxError as e:
            error.extend(e)

        error.propagate()
        return result

    @property
    def specifications(self) -> Dict[str, Specification]:
        return {
            spec_node.spec.f_package_declaration.f_identifier.text: spec_node.spec
            for spec_node in self.__specifications.values()
        }

    def __evaluate_specification(
        self, error: RecordFluxError, spec: Specification, filename: Path
    ) -> None:
        handlers = {
            "SequenceTypeDef": create_sequence,
            "ModularTypeDef": create_modular,
            "RangeTypeDef": create_range,
            "MessageTypeDef": create_message,
            "NullMessageTypeDef": create_null_message,
            "TypeDerivationDef": create_derived_message,
            "EnumerationTypeDef": create_enumeration,
        }
        log.info("Processing %s", spec.f_package_declaration.f_identifier.text)
        package_id = create_id(spec.f_package_declaration.f_identifier, filename)

        for t in spec.f_package_declaration.f_declarations:
            if isinstance(t, TypeDecl):
                identifier = model.qualified_type_identifier(
                    create_id(t.f_identifier, filename), package_id
                )
                try:
                    new_type = handlers[t.f_definition.kind_name](
                        identifier,
                        t.f_definition,
                        self.__types,
                        self.skip_verification,
                        self.__cache,
                        filename,
                    )
                    self.__types.append(new_type)
                    error.extend(new_type.error)
                except RecordFluxError as e:
                    error.extend(e)
            elif isinstance(t, RefinementDecl):
                try:
                    new_type = create_refinement(t, package_id, self.__types, filename)
                    self.__types.append(new_type)
                    error.extend(new_type.error)
                except RecordFluxError as e:
                    error.extend(e)
            elif isinstance(t, SessionDecl):
                try:
                    new_session = create_session(t, package_id, filename, self.__types)
                    self.__sessions.append(new_session)
                    error.extend(new_session.error)
                except RecordFluxError as e:
                    error.extend(e)
            else:
                raise NotImplementedError(f"Declaration kind {t.kind_name} unsupported")
